from __future__ import annotations
import numpy as np
import typing as tp
import numpy.typing as npt
from . import Q_
from .esbmtk_base import esbmtkBase  # , Reservoir, Flux, Source, Sink
from .solver import get_l_mass

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]

np.set_printoptions(precision=4)
# from .connections import ConnnectionGroup

if tp.TYPE_CHECKING:
    from esbmtk import Flux, Model
#     from .esbmtk import Source, Reservoir, Sink, Flux, Model
#     from .extended_classes import GasReservoir, ReservoirGroup
#     from .connections import ConnectionGroup


class Process(esbmtkBase):
    """This class defines template for proycess which acts on one or more
    reservoir flux combinations. To use it, you need to create an
    subclass which defines the actual process implementation in their
    call method. See 'PassiveFlux as example'

    """

    from .esbmtk import Source, Reservoir, Sink

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        self.__defaultnames__()  # default kwargs names
        self.__initerrormessages__()  # default error messages
        self.bem.update({"rate": "a string"})
        self.bem.update({"scale": "Number or quantity"})
        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values

        self.__postinit__()  # do some housekeeping
        self.parent = self.register
        self.__register_name_new__()

    def __postinit__(self) -> None:
        """Do some housekeeping for the process class"""

        from esbmtk import Reservoir

        # legacy name aliases
        self.n: str = self.name  # display name of species
        if "reservoir" in self.kwargs:
            self.r: Reservoir = self.reservoir

        self.mo: Model = self.model
        self.f: Flux = self.flux

        # self.rm0: float = self.r.m[0]  # the initial reservoir mass

        if "reservoir" in self.kwargs and isinstance(self.r, Reservoir):
            self.direction: dict[Flux, int] = self.r.lio[self.f]

        self.delta = self.kwargs["delta"] if "delta" in self.kwargs else "None"
        self.__misc_init__()

    def __delayed_init__(self) -> None:
        """
        Initialize stuff which is only known after the entire model has been defined.
        This will be executed just before running the model. You need to add the following
        two lines somewhere in the init procedure (preferably by redefining __misc_init__)
        and redefine __delayed_init__ to actually execute any code below

        # this process requires delayed init
        self.mo.lto.append(self)

        """

        pass

    def __misc_init__(self) -> None:
        """This is just a place holder method which will be called by default
        in __post_init__() This can be overloaded to add additional
        code to the init procedure without the need to redefine
        init. This useful for processes which only define a call method.

        """

        pass

    def __defaultnames__(self) -> None:
        """Set up the default names and dicts for the process class. This
        allows us to extend these values without modifying the entire init process

        """

        from esbmtk import (
            Reservoir,
            ReservoirGroup,
            Source,
            Sink,
            GasReservoir,
            Flux,
            ConnectionGroup,
            Model,
        )

        # provide a dict of known keywords and types
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "reservoir": ["None", (str, Reservoir, Source, Sink, GasReservoir)],
            "flux": ["None", (str, Flux)],
            "ref_flux": ["None", (str, Flux)],
            "rate": [0, (int, float, np.ndarray)],
            "delta": ["None", (int, float, np.ndarray, str)],
            "lt": ["None", (Flux)],
            "alpha": [0, (int, float, np.ndarray)],
            "scale": [1, (int, float, np.ndarray)],
            "ref_reservoirs": ["None", (Flux, Reservoir, GasReservoir, list, str)],
            "model": ["None", (str, Model)],
            "source": ["None", (str, Source, Reservoir, GasReservoir)],
            "register": [
                "None",
                (
                    str,
                    ConnectionGroup,
                    Reservoir,
                    ReservoirGroup,
                    GasReservoir,
                    Flux,
                    Model,
                ),
            ],
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name", "register"]

    def __register__(self, reservoir, flux: Flux) -> None:
        """Register the flux/reservoir pair we are acting upon, and register
        the process with the reservoir

        """

        from esbmtk import Reservoir

        # register the reservoir flux combination we are acting on
        self.f: Flux = flux
        self.r: Reservoir = reservoir
        # add this process to the list of processes acting on this reservoir
        reservoir.lop.append(self)
        flux.lop.append(self)
        # Add to model flux list
        reservoir.mo.lop.append(self)

    def show_figure(self, x, y) -> None:
        """Apply the current process to the vector x, and show the result as y.
        The resulting figure will be automatically saved.

        Example::
             process_name.show_figure(x,y)

        """
        pass


class GenericFunction(Process):
    """This Process class creates a GenericFunction instance which is
    typically used with the VirtualReservoir, and
    ExternalCode classes. This class is not user facing,
    please see the ExternalCode class docstring for the
    function template of a user provided function.

    see calc_carbonates in the carbonate chemistry for an example how
    to write a function for this class.

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options

        """

        from esbmtk import ReservoirGroup
        from typing import Callable
        
        self.__defaultnames__()  # default kwargs names

        # list of allowed keywords
        defaults: dict[str, list[any, tuple]] = {
            "function": ["None", (str, Callable)],
            "input_data": ["None", (list, str)],
            "vr_data": ["None", (list, str)],
            "function_params": ["None", (list, str)],
            "model": ["None", (str, Model)],
            "ftype": ["None", (str)],
            "r_s": ["None", (str, ReservoirGroup)],
            "r_g": ["None", (str, ReservoirGroup)],
        }

        self.defaults.update(defaults)
        self.__initialize_keyword_variables__(kwargs)
        # required arguments
        self.lrk: list = ["name", "input_data", "vr_data", "function_params", "model"]

        self.mo = self.model
        self.parent = self.register

        self.__postinit__()  # do some housekeeping

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name_new__()  #

    def __call__(self, i: int) -> None:
        """Here we execute the user supplied function
        Where i = index of the current timestep

        """

        self.function(i, self.input_data, self.vr_data, self.function_params)


class AddSignal(Process):
    """This process adds values to the current flux based on the values provided by
    the signal object. This class is typically invoked through the connector object:

    Example::

     AddSignal(name = "name",
               reservoir = upstream_reservoir_handle,
               flux = flux_to_act_upon,
               lt = flux with lookup values,
     )

    where - the upstream reservoir is the reservoir the process belongs too,
    the flux is the flux to act upon, and lt= contains the flux object we lookup from

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["lt", "flux", "reservoir"])  # new required keywords
        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.parent = self.register
        self.__register_name_new__()

        # defaults
        self.__execute__ = self.__add_with_fi__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    # use this when we do isotopes
    def __add_with_fi__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)

        Note that the signal may also specify a delta value. Thus we
        need to
        1) get the mass of flux + signal
        2) add the delta of flux and signal
        3) calculate the new li and hi

        """

        # add signal mass to flux mass
        r = self.f.species.r
        fm = self.f.m[i]  # flux rate
        fl = self.f.l[i]  # flux rate light isotope
        sm = self.lt.m[i]  # signal mass
        sl = self.lt.l[i]  # signal li

        # get signal delta
        sd = 1000 * ((sm - sl) / sl - r) / r if sm > 0 else 0
        # print(f"Signal delta = {sd}")

        fd = 1000 * ((fm - fl) / fl - r) / r + sd if fm > 0 else sd
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        self.f.fa = np.array([fm, fl])
        # print(f"fa = {self.f.fa}\n")

    def __add_with_fa__(self, i) -> None:
        """same as above but use fa instead of flux data"""

        # add signal mass to flux mass
        r = self.f.species.r
        fm = self.f.fa[0]  # flux rate
        fl = self.f.fa[1]  # flux rate light isotope
        sm = self.lt.m[i]  # signal mass
        sl = self.lt.l[i]  # signal li

        # get signal delta
        sd = 1000 * ((sm - sl) / sl - r) / r

        fd = 1000 * ((fm - fl) / fl - r) / r + sd if fm > 0 else sd
        # set new flux rate
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        self.f.fa = np.array([fm, fl])

    def p_add_signal_fi(data, params, i) -> None:
        r: float = params[0]
        fm: float = data[0][i]  # fm
        fl: float = data[1][i]  # fl
        sm: float = data[2][i]  # sm
        sl: float = data[3][i]  # sd

        # get signal delta
        sd = 1000 * ((sm - sl) / sl - r) / r if sm > 0 else 0
        fd = 1000 * ((fm - fl) / fl - r) / r + sd if fm > 0 else sd
        # set new flux rate
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        data[4][:] = [fm, fl]

    def p_add_signal_fa(data, params, i) -> None:
        r: float = params[0]
        fm: float = data[2][0]
        fl: float = data[2][1]
        sm: float = data[0][i]
        sl: float = data[1][i]

        # get signal delta
        sd = 1000 * ((sm - sl) / sl - r) / r if sm > 0 else 0
        fd = 1000 * ((fm - fl) / fl - r) / r + sd if fm > 0 else sd
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)
        data[2][:] = [fm, fl]


class SaveFluxData(Process):
    """
    This process stores the flux data from each iteration into a vector
    Example::

        SaveFluxData(name = "Name",
                     flux = Flux Handle,
                     )

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this Process"""
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["flux"])  # new required keywords

        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values

        # legacy variables
        self.parent = self.register
        self.__postinit__()  # do some housekeeping
        self.__register_name_new__()

    # setup a placeholder call function
    def __call__(self, i: int):
        self.f[i] = self.f.fa

    def p_save_flux(data, params, i) -> None:
        data[0][i] = data[2][0]
        data[1][i] = data[2][1]


class ScaleFlux(Process):
    """This process scales the mass of a flux (m,l,h) relative to
    another flux. Delta is either taken from the upstream reservoir
    or set to a fixed value.  The scale factor "scale" and flux
    reference must be present when the object is being initalized

    Example::
         ScaleFlux(name = "Name",
                   reservoir = reservoir_handle (upstream or downstream)
                   scale = 1
                   ref_flux = flux we use for scale
                   )

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this Process"""

        from esbmtk import Source

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "scale"])  # new required keywords
        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values

        if "ref_reservoirs" in kwargs:
            self.ref_flux = kwargs["ref_reservoirs"]
        elif "ref_flux" not in kwargs:
            raise ValueError("You need to specify a value for ref_flux")

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.parent = self.register
        self.__register_name_new__()

        """ Decide how to calculate isotopes (if)
        If the connection specifies delta explicitly, this will
        override any settings in the source.
        """
        self.c = 1
        if self.delta != "None":
            # delta explicitly provided
            l = get_l_mass(1, self.delta, self.reservoir.species.element.r)
            self.c = l / (1 - l)
            self.__execute__ = self.__with_fixed_delta__
        elif isinstance(self.source, Source):
            if self.source.delta != "None":
                self.delta = self.source.delta
                l = get_l_mass(1, self.delta, self.reservoir.species.element.r)
                self.c = l / (1 - l)
                self.__execute__ = self.__with_fixed_delta__
            else:
                self.__execute__ = self.__without_isotopes__
        elif self.source.isotopes:
            self.__execute__ = self.__with_isotopes__
        else:
            self.__execute__ = self.__without_isotopes__

    # create stubs
    def __with_source__():
        raise NotImplementedError()

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    def __with_isotopes__(self, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
        model execute method.
        Note that this will use the mass of the flux we use for scaling, but that we will set the
        delta according to reservoir this flux derives from

        """

        # get reference flux
        rm: float = self.ref_flux.fa[0] * self.scale

        # get the target isotope ratio based on upstream delta
        c = self.reservoir.l[i - 1] / (
            self.reservoir.c[i - 1] - self.reservoir.l[i - 1]
        )

        fl: float = rm * c / (c + 1)

        self.flux.fa[:] = [rm, fl]

    def __with_fixed_delta__(self, i: int) -> None:
        """similar to with isotopes, but this time we take the
        isotope value of the source (which has no array, just a fixed
        value"""

        # get reference flux
        rm: float = self.ref_flux.fa[0] * self.scale

        # get the target isotope ratio based on upstream delta
        c = self.c

        fl: float = rm * c / (c + 1)

        self.flux.fa[:] = [rm, fl]

    def __without_isotopes__(self, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
        model execute method.
        Note that this will use the mass of the reference object, but that we will set the
        delta according to the reservoir (or the flux?)

        """

        self.f.fa = np.array(
            [
                self.ref_flux.fa[0] * self.scale,
                0,
            ]
        )

    def p_scale_flux_r(data, params, i) -> None:
        """delta is derived from upstream reservoir"""

        # params
        s: float = params[1]  # scale

        # data
        mf: float = data[6][0] * s  # mass of reference flux
        mr: float = data[3][i - 1]  # mass upstream reservoir
        lr: float = data[4][i - 1]  # li upstream reservoir

        # get the target isotope ratio based on upstream delta
        c = lr / (mr - lr)
        l = mf * c / (c + 1)

        data[5][:] = [mf, l]

    def p_scale_flux_fd(data, params, i) -> None:
        """delta is set to a fixed value"""

        # params
        s: float = params[1]  # scale
        c: float = params[2]  # l/h ratio

        # data
        mf: float = data[6][0] * s  # mass of reference flux

        # get the target isotope ratio based on upstream delta
        l = mf * c / (c + 1)

        data[5][:] = [mf, l]


class Fractionation(Process):
    """This process offsets the isotopic ratio of the flux by a given
       delta value. In other words, we add a fractionation factor

    Example::
         Fractionation(name = "Name",
                       reservoir = upstream_reservoir_handle,
                       flux = flux handle
                       alpha = 12 in permil (e.f)

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this Process"""
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "alpha"])  # new required keywords

        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping

        # alpha is given in permil, but the fractionation routine expects
        # it as 1 + permil, i.e., 70 permil would 1.007

        self.alp = 1 + self.alpha / 1000
        self.mo = self.reservoir.mo
        self.parent = self.register
        self.__register_name_new__()

    def __call__(self, i: int) -> None:
        """
        Set flux isotope masses based on fractionation factor
        relative to reservoir

        """

        # print(f"self.f.m[i] =  {self.f.m[i]}")
        fm = self.f.fa[0]
        if fm != 0:
            rm = self.reservoir.m[i - 1]
            rl = self.reservoir.l[i - 1]
            a = self.alp

            fl = rl * fm / (a * rm + rl - a * rl)
            self.f.fa[:] = [fm, fl]
        else:
            self.f.fa[:] = [
                0,
                0,
            ]

        return

    def p_fractionation(data, params, i) -> None:
        # params
        a: float = params[1]  # alpha

        # data
        fm: float = data[4][0]  # flux mass
        rm: float = data[2][i - 1]  # 4 reservoir mass
        rl: float = data[3][i - 1]  # 4 reservoir light isotope

        fl = rl * fm / (a * rm + rl - a * rl)
        data[4][:] = [fm, fl]


class RateConstant(Process):
    """This is a wrapper for a variety of processes which depend on rate constants
    Please see the below class definitions for details on how to call them
    At present, the following processes are defined



    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this Process"""

        from esbmtk import (
            SeawaterConstants,
            GasReservoir,
            Reservoir,
            Source,
            Sink,
            Q_,
        )
        from typing import Callable

        # Note that self.lkk values also need to be added to the lkk
        # list of the connector object.

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names

        # update the allowed keywords
        defaults: dict[str, list[any, tuple]] = {
            "scale": [1, (int, float, np.ndarray)],
            "k_value": [1, (int, float, np.ndarray)],
            "ref_reservoirs": ["None", (str, list)],
            "reservoir_ref": ["None", (str, Reservoir, GasReservoir)],
            "left": [
                "None",
                (str, list, Reservoir, int, float, np.ndarray, np.ndarray),
            ],
            "right": ["None", (str, list, Reservoir, int, float, np.ndarray)],
            "gas": [
                "None",
                (str, Reservoir, GasReservoir, Source, Sink, np.ndarray, float),
            ],
            "liquid": ["None", (Reservoir, Source, Sink, float)],
            "solubility": ["None", (str, int, float, np.ndarray)],
            "piston_velocity": ["None", (str, int, float, np.ndarray)],
            "water_vapor_pressure": ["None", (str, int, float, np.ndarray)],
            "ref_species": ["None", (str, np.ndarray, float)],
            "seawaterconstants": ["None", (str, SeawaterConstants)],
            "isotopes": [False, (bool)],
            "function_reference": ["None", (str, Callable)],
            "f_0": ["None", (str, Q_, float, int)],
            "pco2_0": ["280 ppm", (str, Q_, float)],
            "ex": [0.2, (int, float)],
            "source": ["None", (str, Source, GasReservoir)],
        }

        self.defaults.update(defaults)
        # new required keywords
        # self.lrk.extend([["reservoir", "atmosphere"], ["scale", "k_value"], "register"])

        self.__initialize_keyword_variables__(kwargs)

        self.parent = self.register
        if "reservoir" in kwargs:
            self.mo = self.reservoir.mo
        elif "gas_reservoir" in kwargs:
            self.mo = self.gas_reservoir

        self.__misc_init__()
        self.__postinit__()  # do some housekeeping
        # legacy variables

        self.__register_name_new__()

        self.c = 1
        if self.reservoir.isotopes or self.isotopes:
            self.__execute__ = self.__with_isotopes__
            if self.delta != "None":
                self.__execute__ = self.__with_fixed_delta__
                l = get_l_mass(1, self.delta, self.reservoir.species.r)
                self.c = l
                # print(f"delta = {self.delta} setting c = {self.c}")
        else:
            self.__execute__ = self.__without_isotopes__

    def __postinit__(self) -> "None":
        self.mo = self.reservoir.mo

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)


class weathering(RateConstant):
    """This process calculates the flux as a function of the upstream
     reservoir concentration C and a constant which describes thet
     strength of relation between the reservoir concentration and
     the flux scaling

     F = f_0 * (scale * C/pco2_0)**ncc

     where C denotes the concentration in the ustream reservoir, k is a
     constant. This process is typically called by the connector
     instance. However you can instantiate it manually as

     weathering(
                       name = "Name",
                       reservoir = upstream_reservoir_handle,
                       reservoir_ref = reference_reservoir (Atmosphere)
                       flux = flux handle,
                       ex = exponent
                       pco2_0 = 280,
                       f_0 = 12 / 17e12
                       Scale =  1000,
                       delta = 0,

    )

    """

    def __misc_init__(self):
        """
        Scale the flux relative to the pco2_0 concentration.
        The delta values are derived from either:

        - the upstream reservoir
        - the upstream source
        - set to a fixed value

        """

        from esbmtk import Source

        if isinstance(self.f_0, str):
            self.f_0: float = Q_(self.f_0).to(self.mo.f_unit).magnitude
        elif isinstance(self.f_0, Q_):
            self.f_0: float = self.f_0.to(self.mo.f_unit).magnitude

        if isinstance(self.pco2_0, str):
            self.pco2_0 = Q_(self.pco2_0).to("ppm").magnitude * 1e-6
        elif isinstance(self.pco2_0, Q_):
            self.pco2_0 = self.pco2_0.magnitude.to("ppm").magnitude * 1e-6

        # if self.delta == "None" and  self.source.isotopes == False:
        #      self.fixed = True
        # delta provided explicitly
        if self.delta != "None":
            self.d = self.delta
            self.isotopes = True
            l = get_l_mass(1, self.d, self.source.species.element.r)
            self.c = l / (1 - l)
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__with_fixed_delta__
            self.fixed = True

        # source is Source and has delta
        elif self.source.isotopes and isinstance(self.source, Source):
            self.isotopes = True
            self.delta = self.source.delta
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__with_fixed_delta__
            self.c = self.source.c
            self.fixed = True

        # source is reservoir
        elif self.source.isotopes:
            self.isotopes = True
            self.func_name = self.p_weathering
            self.__execute__ = self.__with_isotopes__
            self.fixed = False

        # source is source, has no delta, and delta is not provided
        else:
            self.c = 1
            self.isotopes = False
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__without_isotopes__
            self.fixed = True

    def __without_isotopes__(self, i: int) -> None:
        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        self.flux.fa = np.array([f, 0])

    def __with_isotopes__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of reservoir concentration,
        rather than scaling the flux.
        """

        c = self.flux.fa[1] / self.flux.fa[0]
        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        fl = f * c
        self.flux.fa = np.array([f, fl])

    def __with_fixed_delta__(self, i: int) -> None:
        """ """

        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        fl = f * self.c
        self.flux.fa = np.array([f, fl])

    def p_weathering(data, params, i) -> None:
        """delta value depends on upstream reservoir"""
        # params
        s: float = params[1]
        f_0: float = params[2]
        pco2_0: float = params[3]
        ex: float = params[4]

        # data
        cr = data[3][i - 1] / data[2][i - 1]
        c: float = data[1][i - 1]

        f: float = s * f_0 * (c / pco2_0) ** ex
        fl: float = f * cr
        data[0][:] = [f, fl]

    def p_weathering_fd(data, params, i) -> None:
        """delta value is fixed"""
        # params
        s: float = params[1]
        f_0: float = params[2]
        pco2_0: float = params[3]
        ex: float = params[4]
        cr = params[5]

        c: float = data[1][i - 1]
        f: float = s * f_0 * (c / pco2_0) ** ex
        fl = f * cr
        data[0][:] = [f, fl]


class MultiplySignal(Process):
    """This process mulitplies a given flux with the the data in the
    signal.  This class is typically invoked through the connector
    object: Note that this process will not modify the delta value of
    a given flux.  If you needto vary the delta value it is best to
    add a second signal which uses the add signal type.

    Example::
    
       MultiplySignal(name = "name",
               reservoir = upstream_reservoir_handle,
               flux = flux_to_act_upon,
               lt = flux with lookup values,
          )

    where the upstream reservoir is the reservoir the process belongs
    too the flux is the flux to act upon lt= contains the
    flux object we lookup from

    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["lt", "flux", "reservoir"])  # new required keywords
        self.__initialize_keyword_variables__(kwargs)  # initialize keyword values

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.parent = self.register
        self.__register_name_new__()

        # decide whichh call function to use
        # if self.mo.m_type == "both":

        # default
        self.__execute__ = self.__multiply_with_flux_fi__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    # use this when we do isotopes
    def __multiply_with_flux_fi__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)
        """

        # multiply flux mass with signal
        c = self.lt.m[i]
        m = self.f.m[i] * c
        l = self.f.l[i] * c
        # h = self.f.h[i] * c
        # d = self.f.d[i]
        self.flux.fa = np.array([m, l])
        print(f"multiply fa {self.flux.fa}, f. = {self.f.m[i]}, c = {c} ")

    def __multiply_with_flux_fa__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)
        """

        # multiply flux mass with signal
        c = self.lt.m[i]
        m = self.f.fa[0] * c
        l = self.f.fa[1] * c
        self.flux.fa = np.array([m, l])

    def p_multiply_signal_fi(data, params, i) -> None:
        c = data[2][i]
        m = data[0][i] * c
        l = data[1][i] * c

        data[3][:] = [m, l]

    def p_multiply_signal_fa(data, params, i) -> None:
        c = data[0][i]
        m = data[1][0] * c  # m
        l = data[1][1] * c  # l

        data[1][:] = [m, l]


class VarDeltaOut(Process):
    """Unlike a passive flux, this process sets the flux istope ratio
    equal to the isotopic ratio of the reservoir. The
    init and register methods are inherited from the process
    class::
    
     VarDeltaOut(name = "name",
                reservoir = upstream_reservoir_handle,
                flux = flux handle,
                rate = rate,
     )
     
    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this Process"""

        from esbmtk import (
            Reservoir,
            Source,
        )

        # get default names and update list for this Process
        self.__defaultnames__()

        self.lrk.extend(["reservoir", "flux", "register"])  # required keywords

        self.__initialize_keyword_variables__(kwargs)
        self.mo = self.reservoir.mo
        self.parent = self.register
        self.__postinit__()  # do some housekeeping
        self.__register_name_new__()

        # decide which call function to use
        # if self.mo.m_type == "both":
        if self.reservoir.isotopes:
            # print(
            #    f"vardeltaout with isotopes for {self.reservoir.register.name}.{self.reservoir.name}"
            # )
            if isinstance(self.reservoir, Reservoir):
                # print("Using reservoir")
                self.__execute__ = self.__with_isotopes_reservoir__
            elif isinstance(self.reservoir, Source):
                # print("Using Source")
                self.__execute__ = self.__with_isotopes_source__
            else:
                raise ValueError(
                    f"{self.name}, reservoir must be of type Source or Reservoir, not {type(self.reservoir)}"
                )
        else:
            self.__execute__ = self.__without_isotopes__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    def __with_isotopes_reservoir__(self, i: int) -> None:
        """Here we re-balance the flux. This code will be called by the
        apply_flux_modifier method of a reservoir which itself is
        called by the model execute method
        """

        m: float = self.flux.m[i]
        if m != 0:
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            self.flux.fa = [m, m * c]

    def p_vardeltaout(data, params, i) -> None:
        # concentration times scale factor

        mf: float = data[0][i - 1]  # flux mass
        mr: float = data[1][i - 1]  # reservoir mass
        lr: float = data[2][i - 1]  # reservoir l

        c = lr / (mr - lr)
        data[3][:] = [mf, mf * c]

    def __with_isotopes_source__(self, i: int) -> None:
        """If the source of the flux is a source, there is only a single delta value.
        Changes to the flux delta are applied through the Signal class.
        """

        m: float = self.flux.m[i]
        if m != 0:
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            self.flux.fa = [m, m * c]

    def __without_isotopes__(self, i: int) -> None:
        """Here we re-balance the flux. This code will be called by the
        apply_flux_modifier method of a reservoir which itself is
        called by the model execute method
        """

        raise NotImplementedError("vardeltaout w/o isotopes is not defined")


class ScaleRelativeToMass(RateConstant):
    """This process scales the flux as a function of the upstream
    reservoir Mass M and a constant which describes the
    strength of relation between the reservoir mass and
    the flux scaling

    F = F0 *  M * k

    where M denotes the mass in the ustream reservoir, k is a
    constant and F0 is the initial unscaled flux. This process is
    typically called by the connector instance. However you can
    instantiate it manually as

    Note that we scale the flux, rather than compute the flux!

    This is faster than setting a new flux, computing the isotope
    ratio and setting delta. So you either have to set the initial
    flux F0 to 1, or calculate the scale accordingly::

     ScaleRelativeToMass(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       Scale =  1000,
     )

    """

    def __without_isotopes__(self, i: int) -> None:
        m: float = self.reservoir.m[i - 1] * self.scale
        self.flux.fa = [m, 0, 0, 0]

    def __with_isotopes__(self, i: int) -> None:
        """
        this will be called by the Model.run() method

        """
        m: float = self.reservoir.m[i - 1] * self.scale
        # d: float = self.reservoir.d[i - 1]
        c = self.reservoir.l[i - 1] / (
            self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
        )
        l = m * c / (c + 1)
        self.flux.fa[:] = [m, l]

    def p_scale_relative_to_mass(data, params, i) -> None:
        # concentration times scale factor
        # params
        s: float = params[1]  # scale factor

        # data
        rm = data[2][i - 1]
        rl = data[3][i - 1]
        # rd = data[6][i - 1]

        m: float = rm * s  # new flux
        c: float = rl / (rm - rl)
        l: float = m * c / (c + 1)
        # d: float = data[1][i - 1]  # flux delta

        data[4][:] = [m, l]


class GasExchange(RateConstant):
    """
    Example::
    
      GasExchange(
          gas =GasReservoir, #
          liquid = Reservoir, #,
          ref_species = array of concentrations #
          solubility=Atmosphere.swc.SA_co2 [mol/(m^3 atm)],
          area = area, # m^2
          piston_velocity = m/year
          seawaterconstants = Ocean.swc
          water_vapor_pressure=Ocean.swc.p_H2O,
      )

    """

    # redefine misc_init which is being called by post-init
    def __misc_init__(self):
        """Set up input variables"""

        self.p_H2O = self.seawaterconstants.p_H2O
        self.a_dg = self.seawaterconstants.a_dg
        self.a_db = self.seawaterconstants.a_db
        self.a_u = self.seawaterconstants.a_u
        self.rvalue = self.liquid.sp.r
        self.volume = self.gas.volume

    def ode(self) -> None:
        return self.scale * (
            self.gas.c * (1 - self.p_H2O) * self.solubility - self.ref_species * 1000
        )  # area in m^2  # Atmosphere  # p_H2O  # SA_co2 = mol/(m^3 atm)  # [CO2]aq mol

    def __without_isotopes__(self, i: int) -> None:
        # set flux
        # note that the sink delta is co2aq as returned by the carbonate VR
        # this equation is for mmol but esbmtk uses mol, so we need to
        # multiply by 1E3

        f = self.scale * (  # area in m^2
            self.gas.c[i - 1]  # Atmosphere
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility  # SA_co2 = mol/(m^3 atm)
            - self.ref_species[i - 1] * 1000  # [CO2]aq mol
        )
        # print(self.gas.c[i - 1])

        # changes in the mass of CO2 also affect changes in the total mass
        # of the atmosphere. So we need to update the reservoir volume
        # variable which we use the store the total atmospheric mass
        """The solver will use f and f12 to update mass, light
        isotope and concentration values in the Gas
        reservoir. However, Gas reservoirs track total gas pressure
        in the volume variable. This will not be updated by the
        solver. So we need to do it ourseleves

        """
        # self.gas.v[i] = self.gas.v[i - 1] + f * self.gas.mo.dt
        self.flux.fa[:] = [f, 1]

    def __with_isotopes__(self, i: int) -> None:
        """
        In the following I assume near neutral pH between 7 and 9, so that
        the isotopic composition of HCO3- is approximately equal to the isotopic
        ratio of DIC. The isotopic ratio of [CO2]aq can then be obtained from DIC via
        swc.e_db (swc.a_db)

        The fractionation factor subscripts denote the following:

        g = gaseous
        d = dissolved
        b = bicarbonate ion
        c = carbonate ion

        a_db is thus the fractionation factor between dissolved CO2aq and HCO3-
        and a_dg between CO2aq and CO2g

        ref_species = DIC.cs.CO2aq
        liquid.m = DIC.m (used to get the isotope ratio)

        gas.m = mass of CO2
        gas.l = mass of 12CO2
        gas.c = CO2 concentration
        gas.v = total mass of atmosphere

        """

        # equilibrium concentration of CO2 in water based on pCO2
        eco2_at = (
            self.gas.c[i - 1]  # p Atmosphere
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility
        )
        # equilibrium concentration of CO2 in water based on CO2aq
        eco2_aq = self.ref_species[i - 1] * 1000

        #  flux
        f = self.scale * (eco2_at - eco2_aq)

        c13g = 1 - self.gas.l[i - 1] / self.gas.m[i - 1]
        c13aq = 1 - self.liquid.l[i - 1] / self.liquid.m[i - 1]
        # get 13C CO2 equlibrium concentration  CO2 in water based on pCO2
        eco2_at_13 = (
            self.gas.c[i - 1]
            * c13g
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility
            * self.a_dg
        )

        # get 13C equilibrium  CO2 in water based on DIC m & l
        eco2_aq_13 = self.a_db * eco2_aq * c13aq

        # 13C flux
        f13 = self.scale * self.a_u * (eco2_at_13 - eco2_aq_13)
        f12 = f - f13

        self.flux.fa = [f, f12]
        """The solver will use f and f12 to update mass, light
        isotope and concentration values in the Gas
        reservoir. However, Gas reservoirs track total gas pressure
        in the volume variable. This will not be updated by the
        solver. So we need to do it ourseleves

        not working, see note below
        """
        self.gas.v[i] = self.gas.v[i - 1] + f * self.gas.mo.dt
        return [f, f12]
        # print()

    def __postinit__(self) -> None:
        """Do some housekeeping for the process class"""

        # legacy name aliases
        self.n: str = self.name  # display name of species
        self.r = self.liquid
        self.reservoir = self.liquid


class ScaleRelativeToConcentration(RateConstant):
    """This process calculates the flux as a function of the upstream
    reservoir concentration C and a constant which describes thet
    strength of relation between the reservoir concentration and
    the flux scaling:

    F = C * k

    where C denotes the concentration in the ustream reservoir, k is a
    constant. This process is typically called by the connector
    instance. However you can instantiate it manually as::

     ScaleRelativeToConcentration(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       Scale =  1000,
                       delta = "None",
     )

    If delta is given the fluxes uses a ficed delta, If not it uses
    the upestream delta

    """

    def __without_isotopes__(self, i: int) -> None:
        m: float = self.reservoir.m[i - 1]
        if m > 0:  # otherwise there is no flux
            # convert to concentration
            c = m / self.reservoir.volume
            f = c * self.scale
            self.flux.fa = [
                f,
                0,
            ]

    def __with_isotopes__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of self.reservoir concentration,
        rather than scaling the flux.
        """

        rc: float = self.reservoir.c[i - 1]
        if rc > 0:  # otherwise there is no flux
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            m = rc * self.scale
            l = m * c / (c + 1)
            self.f.fa = [m, l]

    def __with_fixed_delta__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of self.reservoir concentration,
        rather than scaling the flux.
        """

        rc: float = self.reservoir.c[i - 1]
        if rc > 0:  # otherwise there is no flux
            c = self.c
            m = rc * self.scale
            l = m * c / (c + 1)
            self.f.fa = [m, l]

    def p_scale_relative_to_concentration(data, params, i) -> None:
        """delta depends on upstream delta"""
        # data
        rm: float = data[0][i - 1]  # res mass
        if rm > 0:
            # params
            s: float = params[0]  # scale factor
            v: float = params[1]  # res volume
            fm: float = rm / v * s
            rl: float = data[1][i - 1]  # res li

            c: float = rl / (rm - rl)
            fl: float = fm * c / (c + 1)
            data[2][:] = [fm, fl]

        else:
            data[2][:] = [0.0, 0.0]

    def p_scale_relative_to_concentration_fd(data, params, i) -> None:
        """delta depends fixed value"""
        # data
        rm: float = data[0][i - 1]  # res mass

        if rm > 0:
            # params
            s: float = params[1]  # scale factor
            v: float = params[2]  # res volume
            fm: float = rm / v * s
            c: float = params[3]  # isotope ratio
            fl: float = fm * c / (c + 1)
            data[2][:] = [fm, fl]

        else:
            data[2][:] = [0.0, 0.0]



"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.

     This module defines some shared methods

"""
from __future__ import annotations
import time
import numpy as np
import numpy.typing as npt

import typing as tp

if tp.TYPE_CHECKING:
    from .esbmtk import Species

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


class KeywordError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class MissingKeywordError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class InputError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class FluxSpecificationError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class SpeciesMolweightError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class input_parsing(object):
    """Provides various routines to parse and process keyword
    arguments.  All derived classes need to declare the allowed
    keyword arguments, their defualt values and the type in the
    following format:

    defaults = {"key": [value, (allowed instances)]

    the recommended sequence is to first set default values via
    __register_variable_names__()

    __update_dict_entries__(defaults,kwargs) will  compare the provided kwargs against
    this data, and upon succesful parsing update the default dict
    with the new values
    """

    def __init__(self):
        raise NotImplementedError("input parsing has no instance!")

    def __initialize_keyword_variables__(self, kwargs) -> None:
        """check, register and update keyword variables"""

        # import copy

        self.update = False
        # self.defaults_copy = copy.deepcopy(self.defaults)
        # self.defaults_copy = {**self.defaults} # probably the wrong place.
        self.__check_mandatory_keywords__(self.lrk, kwargs)
        self.__register_variable_names__(self.defaults, kwargs)
        self.__update_dict_entries__(self.defaults, kwargs)
        self.update = True

    def __check_mandatory_keywords__(self, lrk: list, kwargs: dict) -> None:
        """Verify that all elements of lrk have a corresponding key in
        kwargs.  If not, print error message"""

        for key in lrk:
            if isinstance(key, list):
                has_key = sum(k in kwargs and kwargs[k] != "None" for k in key)
                if has_key != 1:
                    raise ValueError(f"give only one of {key}")
            elif key not in kwargs:
                raise MissingKeywordError(f"{key} is a mandatory keyword")

    def __register_variable_names__(
        self,
        defaults: dict[str, list[any, tuple]],
        kwargs: dict,
    ) -> None:
        """Register the key value[0] pairs as local instance variables.
        We register them with their actual variable name and as _variable_name
        in case we use setter and getter methods.
        to avoid name conflicts.
        """
        for key, value in defaults.items():
            setattr(self, f"_{key}", value[0])
            setattr(self, key, value[0])

        # save kwargs dict
        self.kwargs: dict = kwargs

    def __update_dict_entries__(
        self,
        defaults: dict[str, list[any, tuple]],
        kwargs: dict[str, list],
    ) -> None:
        """This function compares the kwargs dictionary with the defaults
        dictionary. If the kwargs key cannot be found, raise an
        error. Otherwise test that the value is of the correct type. If
        yes, update the defaults dictionary with the new value.

        defaults = {"key": [value, (allowed instances)]
        kwargs = {"key": value

        Note that this function assumes that all defaults have been registered
        with the instance via __register_variable_names__()
        """
        for key, value in kwargs.items():
            if key not in defaults:
                raise KeywordError(f"{key} is not a valid keyword")

            if not isinstance(value, defaults[key][1]):
                raise InputError(
                    f"{value} for {key} must be of type {defaults[key][1]}, not {type(value)}"
                )

            defaults[key][0] = value  # update defaults dictionary
            setattr(self, key, value)  # update instance variables
            setattr(self, f"_{key}", value)

    def __register_name_new__(self) -> None:
        """if self.parent is set, register self as attribute of self.parent,
        and set full name to parent.full-name + self.name
        if self.parent == "None", full_name = name
        """

        if self.parent == "None":
            self.full_name = self.name
            reg = self
        else:
            self.full_name = f"{self.parent.full_name}.{self.name}"
            reg = self.parent.model
            # self.full_name, reg.lmo = self.__test_and_resolve_duplicates__(self.full_name, reg.lmo)
            if self.full_name in reg.lmo:
                print("\n -------------- Warning ------------- \n")
                print(f"\t {self.full_name} is a duplicate name in reg.lmo")
                print("\n ---------------------- ------------- \n")
                # raise NameError(f"{self.full_name} is a duplicate name in reg.lmo")
            # register with model
            reg.lmo.append(self.full_name)
            reg.lmo2.append(self)
            reg.dmo.update({self.full_name: self})
            setattr(self.parent, self.name, self)
            self.kwargs["full_name"] = self.full_name
        self.reg_time = time.monotonic()

    def __test_and_resolve_duplicates__(self, name, lmo):
        if name in lmo:
            print(f"\n Warning, {name} is a duplicate, trying with {name}_1\n")
            name = name + "_1"
            name, lmo = self.__test_and_resolve_duplicates__(name, lmo)
        else:
            lmo.append(name)

        return name, lmo


class esbmtkBase(input_parsing):
    """The esbmtk base class template. This class handles keyword
    arguments, name registration and other common tasks

    Useful methods in this class:

    define required keywords in lrk dict:
       self.lrk: list = ["name"]

    define allowed type per keyword in lkk dict:
       self.defaults: dict[str, list[any, tuple]] = {
                                  "name": ["None", (str)],
                                  "model": ["None",(str, Model)],
                                  "salinity": [35, (int, float)], # int or float
                                  }

    parse and register all keywords with the instance
    self.__initialize_keyword_variables__(kwargs)

    register the instance
    self.__register_name_new__ ()

    """

    def __init__(self) -> None:
        raise NotImplementedError

    def __repr__(self, log=0) -> str:
        """Print the basic parameters for this class when called via the print method"""
        from esbmtk import Q_

        m: str = ""

        # suppress output during object initialization
        tdiff = time.monotonic() - self.reg_time

        # do not echo input unless explicitly requestted

        m = f"{self.__class__.__name__}(\n"
        for k, v in self.kwargs.items():
            if not isinstance({k}, esbmtkBase):
                # check if this is not another esbmtk object
                if "esbmtk" in str(type(v)):
                    m = f"{m}    {k} = {v.name},\n"
                elif isinstance(v, str):
                    m = f"{m}    {k} = '{v}',\n"
                elif isinstance(v, Q_):
                    m = f"{m}    {k} = '{v}',\n"
                elif isinstance(v, (list, np.ndarray)):
                    m = f"{m}    {k} = '{v[:3]}',\n"
                else:
                    m = f"{m}    {k} = {v},\n"

        m = "" if log == 0 and tdiff < 1 else f"{m})"
        return m

    def __str__(self, kwargs=None):
        """Print the basic parameters for this class when called via the print method
        Optional arguments

        indent :int = 0 printing offset

        """
        if kwargs is None:
            kwargs = {}
        from esbmtk import Q_

        m: str = ""
        off: str = "  "

        ind: str = kwargs["indent"] * " " if "indent" in kwargs else ""
        index = int(kwargs["index"]) if "index" in kwargs else -2
        m = f"{ind}{self.name} ({self.__class__.__name__})\n"
        for k, v in self.kwargs.items():
            if not isinstance({k}, esbmtkBase):
                # check if this is not another esbmtk object
                if "esbmtk" in str(type(v)):
                    pass
                elif isinstance(v, str) and k != "name":
                    m = f"{m}{ind}{off}{k} = {v}\n"
                elif isinstance(v, Q_):
                    m = f"{m}{ind}{off}{k} = {v}\n"
                elif isinstance(v, np.ndarray):
                    m = f"{m}{ind}{off}{k}[{index}] = {v[index]:.2e}\n"
                elif k != "name":
                    m = f"{m}{ind}{off}{k} = {v}\n"

        return m

    def __lt__(self, other) -> None:
        """This is needed for sorting with sorted()"""

        return self.n < other.n

    def __gt__(self, other) -> None:
        """This is needed for sorting with sorted()"""

        return self.n > other.n

    def info(self, **kwargs) -> None:
        """Show an overview of the object properties.
        Optional arguments are

        indent :int = 0 indentation

        """

        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = " " * indent

        # print basic data bout this object
        print(f"{ind}{self.__str__(kwargs)}")

    def __aux_inits__(self) -> None:
        """Aux initialization code. Not normally used"""

        pass

    def ensure_q(self, arg):
        """Test that a given input argument is a quantity. If not convert
        into quantity
        """
        from esbmtk import Q_

        if isinstance(arg, Q_):
            pass
        elif isinstance(arg, str):
            arg = Q_(arg)
        else:
            raise InputError(f"{arg} must be string or Quantity, not {type(arg)}")

        return arg

    def help(self) -> None:
        """Show all keywords, their fdefault values and allowed types."""
        print(f"\n{self.full_name} has the following keywords:\n")
        for k, v in self.defaults_copy.items():
            print(f"{k} defaults to {v[0]}, allowed types = {v[1]}")
        print()
        print("The following keywords are mandatory:")
        for kw in self.lrk:
            print(f"{kw}")

    def set_flux(self, mass: str, time: str, substance: Species):
        """
        set_flux converts() a flux rate that is specified as rate, time, substance
        so that it matches the correct model units (i.e., kg/s or mol/s)

        Example:

        .. code-block:: python

           M.set_flux("12 Tmol", "year", M.C)

        if model mass units are in mol, no change will be made 
        if model mass units are in kg, the above will return kg C/a (and vice verso)

        :param mass: e.g., "12 Tmol"
        :param time: e.g., "year"
        :param substance: e.g., Species Instance e.g., M.PO4

        :returns: mol/year or g/year

        :raises: FluxSpecificationError
        :raises: SpeciesMolweightError
    """
        from esbmtk import Q_, ureg

        if substance.m_weight > 0:
            mass = Q_(mass)
            g_per_mol = ureg("g/mol")
            if mass.is_compatible_with("g") or mass.is_compatible_with("mol"):
                r = mass.to(
                    substance.mo.m_unit,  # mol
                    "chemistry",  # context
                    mw=substance.m_weight * g_per_mol,  # g/mol
                )
            else:
                message = f"no known conversion for {mass}, and {substance}"
                raise FluxSpecificationError(message)
        else:
            message = f"no mol weight defintion for {substance.full_name}"
            raise SpeciesMolweightError(message)

        return r / ureg(time)



from __future__ import annotations
import typing as tp
import numpy as np
import numpy.typing as npt

if tp.TYPE_CHECKING:
    from esbmtk import Model

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


def set_tolerance(v: float, rtol: float) -> float:
    """Set tolerance values for solve_ivp. See the solve_ivp doumentation

    :param v: value if inititial condition
    :param rtol: relative tolerance
    :return atol: absolute tolerance

    """
    if v > 0.0:
        atol = rtol * abs(v) / 10.0
    else:
        atol = 1e-6
    return atol


def sort_fluxes_by_type(M) -> dict:
    """Sort all model fluxes in to two lists, one containig all constant fluxes,
    and one containing all computed fluxes (i.e. those that require a function).
    All others, will be collected in the 'other' list.

    :param M: Model handle

    :returns: a dictionary of lists containhg the respective flux/connection types
    """
    f_dict = {
        "variable": [],
        "ref": [],
        "fixed": [],
        "other": [],
    }

    for f in M.lof:
        if f.register.ctype == "weathering".lower():
            f_dict["variable"].append(f)
        elif f.register.ctype == "scale_with_concentration".lower():
            f_dict["other"].append(f)
        elif f.register.ctype == "scale_with_flux".lower():
            f_dict["ref"].append(f)
        elif f.register.ctype == "regular".lower():
            f_dict["fixed"].append(f)
        else:
            raise ValueError(
                f"{f.register.ctype} is not implmented in sort_fluxes_by_type"
            )

    return f_dict


def get_types(M: Model, rtol: float) -> tuple(NDArrayFloat, NDArrayFloat, dict, dict):
    """Create the data that is needed to assemble the equation matrix

    :param M: Model hande
    :param rtol: relative tolerance,

    :return atol: NDArrayFloat of absolute tolerances
    :return IC: NDArrayFloat of initial conditions and constant/computed Fluxes
    :return IC_i: dict of reservoir/flux index positions
    :return f_dict: dictionary of flux types
    """
    from esbmtk import Source, Sink

    atol = list()  # list of tolerances
    IC = list()  # IC = r1.l, r1.c, r2.l, r3.l etc)
    IC_i = dict()  # IC_i[r] = index position of r in IC
    M.num_ic = dict()

    # get initial conditions tolerances and indexes for each reservoir
    i = 0
    for r in M.lic:
        IC.append(r.c[0])
        atol.append(set_tolerance(r.c[0], rtol))
        IC_i[r.full_name] = i
        i = i + 1
        if r.isotopes:
            IC.append(r.l[0])
            atol.append(set_tolerance(r.l[0], rtol))
            IC_i[f"{r.full_name}_i"] = i
            i = i + 1

    M.num_ic["reservoirs"] = i

    # get fluxes sorted by connection type and enter them in list of
    # initial conditions
    f_dict = sort_fluxes_by_type(M)

    types_to_append = ["fixed"]
    for tta in types_to_append:
        if tta in f_dict:
            for f in f_dict[tta]:
                # print(f"setting {f.full_name} with {f.rate}")
                IC.append(f.rate)  # add constant to inittial conditions
                atol.append(set_tolerance(r.m[-1], rtol))
                if isinstance(f.register.source, Source):
                    IC_i[f.register.source.full_name] = i
                elif isinstance(f.register.sink, Sink):
                    IC_i[f.register.sink.full_name] = i
                else:
                    IC_i[f.full_name] = i

                i = i + 1
                M.num_ic[tta] = i

    return np.asarray(atol, dtype=float), np.asarray(IC, dtype=float), IC_i, f_dict


def get_matrix_coordinates(
    source,
    sink,
    IC_i,
    scale: float,
    c,
    bypass=False,
) -> None:
    """Set matrix coefficients for source and sinks.

    :param C: Coefficient Matrix
    :param row: row index
    :param col: cold index
    :param scale: value
    :param bypass: whether to bypass the sink term
    """
    from esbmtk import Source, Sink

    coords: list[list] = list()

    if isinstance(c.source, Source):
        row = IC_i[sink]  # this will set the row
        col = IC_i[source]
        coords.append([row, col, scale])
    elif isinstance(c.sink, Sink):
        row = IC_i[source]  # this will set the row
        col = IC_i[sink]
        coords.append([row, col, -scale])
    else:
        row = IC_i[source]  # this will set the row
        col = IC_i[source]
        coords.append([row, col, -scale])

        if not bypass:
            row = IC_i[sink]  # this will set the row
            col = IC_i[source]
            coords.append([row, col, +scale])

    return coords


def set_matrix_coefficients(C, coords):
    """Set matrix coefficients. If values equals zero, set with scale.
    If value is unequal zero, add to existing value

    :param C: Matrix
    :param coords: Matrix coordinates and values [[0,1,1e4],[1,0,1e3]] etc,

    """
    for c in coords:
        # print(f"row = {c[0]}, col = {c[1]}, val = {c[2]:.2e}")
        if C[c[0], c[1]] == 0:
            C[c[0], c[1]] = c[2]
        else:
            # print(f"C before = {C}")
            # print(
            #     f"current = {C[c[0], c[1]]:.2e}, new = {c[2]:.2e}, future = {C[c[0], c[1]] - c[2]:.2e}"
            # )
            C[c[0], c[1]] += c[2]
            # print(f"C after = {C}\n")
        # print(C)


def build_matrix(M: Model, IC: NDArrayFloat, IC_i: dict, f_dict: dict):
    """Create the initial co-efficient Matrix C[rows, cols]), where
    rows = number of reservoirs
    cols = number of initial condition + fixed an computed fluxes

    :param M: Model handle
    :param IC: array of initial conditions and fixed parameters
    :param IC_i: dict of initial condtions index positions
    :param f_dict: dict of flux_types

    :return C: coefficient matrix
    """
    # create initial coefficient matrix
    rows = len(M.lor)
    cols = len(IC)
    C = np.zeros((rows, cols), dtype=float)

    # loop over reservoirs
    for f in M.lof:
        c = f.register  # get connection

        if c.ctype == "scale_with_flux":
            source = c.ref_flux.register.source.full_name
            sink = c.sink.full_name
            scale = c.ref_flux.register.scale * c.scale
            print(f"source = {source}")
            print(f"sink = {sink}")
        else:
            source = c.source.full_name
            sink = c.sink.full_name
            scale = c.scale

        coords = get_matrix_coordinates(
            source,
            sink,
            IC_i,
            scale,
            c,
        )
        set_matrix_coefficients(C, coords)
    return C



from __future__ import annotations
import typing as tp
import numpy as np
import numpy.typing as npt

if tp.TYPE_CHECKING:
    from .esbmtk import ReservoirGroup

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


def carbonate_system_1_pp(rg: ReservoirGroup) -> None:
    """Calculates and returns various carbonate species based on previously calculated
    Hplus, TA, and DIC concentrations.

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    :param rg: A reservoirgroup object with initialized carbonate system
    """

    k1 = rg.swc.K1  # K1
    k2 = rg.swc.K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    boron = rg.swc.boron  # boron
    hplus = rg.Hplus.c
    dic = rg.DIC.c
    ta = rg.TA.c

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    rg.CA: float = ta + fg
    rg.HCO3: float = dic / (1 + (hplus / k1) + (k2 / hplus))
    rg.CO3: float = (rg.CA - rg.HCO3) / 2
    rg.omega = rg.swc.ca2 * rg.CO3 / rg.swc.Ksp
    rg.pH: float = -np.log10(hplus)
    rg.CO2aq: float = dic / (1 + (k1 / hplus) + (k1 * k2 / (hplus**2)))


def carbonate_system_2_pp(
    rg: ReservoirGroup,  # 2 Reservoir handle
    Bm: float,  # 3 CaCO3 export flux as DIC
) -> None:
    """Calculates and returns the fraction of the carbonate rain that is
    dissolved an returned back into the ocean. This functions returns:

    DIC_burial, DIC_burial_l, Hplus, zsnow

    Additionally, it calculates  the following critical depth intervals:

    zsat: top of lysocline
    zcc: carbonate compensation depth

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    """

    # Parameters
    k1 = rg.swc.K1  # K1
    k2 = rg.swc.K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    boron = rg.swc.boron  # boron
    hplus = rg.Hplus.c
    ta = rg.TA.c
    dic = rg.DIC.c
    p = rg.cs.function_input_data
    ksp0 = p[7]
    ca2 = rg.swc.ca2  # Ca2+
    kc = p[8]
    AD = p[9]
    zsat0 = int(abs(p[10]))
    zsat_min = int(abs(p[13]))
    zmax = int(abs(p[14]))

    # calc carbonate alkalinity based t-1
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    rg.CA: float = ta + fg
    rg.HCO3: float = dic / (1 + (hplus / k1) + (k2 / hplus))

    rg.CO3: float = (rg.CA - rg.HCO3) / 2
    rg.CO3[rg.CO3 < 0] = 0
    rg.CO2aq = dic - rg.CO3 - rg.HCO3
    rg.omega = rg.swc.ca2 * rg.CO3 / rg.swc.Ksp
    rg.pH: float = -np.log10(hplus)
    rg.zsat = np.clip(zsat0 * np.log(ca2 * rg.CO3 / ksp0), zsat_min, zmax)
    rg.zcc = zsat0 * np.log(Bm * ca2 / (ksp0 * AD * kc) + ca2 * rg.CO3 / ksp0)  # eq3
    # rg.zsnow = rg.zsnow.c



from __future__ import annotations
from pandas import DataFrame
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
import pandas as pd
import logging
import os
import math
import copy as cp
import typing as tp

if tp.TYPE_CHECKING:
    from esbmtk import Connection, Model

from .esbmtk_base import esbmtkBase
from .esbmtk import ReservoirBase, Reservoir

from .solver import (
    get_imass,
    get_delta,
    get_l_mass,
)

from .utility_functions import (
    get_string_between_brackets,
)

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


class ReservoirGroupError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class SourceSinkGroupError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class FluxError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class SignalError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class DataFieldError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class ESBMTKFunctionError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class ExternalDataError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class GasResrvoirError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class ReservoirGroup(esbmtkBase):
    """This class allows the creation of a group of reservoirs which share
    a common volume, and potentially connections. E.g., if we have twoy
    reservoir groups with the same reservoirs, and we connect them
    with a flux, this flux will apply to all reservoirs in this group.

    A typical examples might be ocean water which comprises several
    species.  A reservoir group like ShallowOcean will then contain
    sub-reservoirs like DIC in the form of ShallowOcean.DIC

    Example::

        ReservoirGroup(name = "ShallowOcean",        # Name of reservoir group
                    volume/geometry = "1E5 l",       # see below
                    delta   = {DIC:0, TA:0, PO4:0]  # dict of delta values
                    mass/concentration = {DIC:"1 unit", TA: "1 unit"}
                    plot = {DIC:"yes", TA:"yes"}  defaults to yes
                    isotopes = {DIC: True/False} see Reservoir class for details
                    seawater_parameters = dict, optional, see below
                    carbonate_system= False, see below
                    register= model handle, required
               )

    Notes: The subreservoirs are derived from the keys in the concentration or mass
           dictionary. Toward this end, the keys must be valid species handles and
           -- not species names -- !

    Connecting two reservoir groups requires that the names in both
    group match, or that you specify a dictionary which delineates the
    matching.

    Most parameters are passed on to the Reservoir class. See the reservoir class
    documentation for details

    The geometry keyword specifies the upper depth interval, the lower
    depth interval, and the fraction of the total ocean area inhabited by the reservoir

    If the geometry parameter is supplied, the following instance variables will be
    computed:

     - self.volume: in model units (usually liter)
     - self.area: surface area in m^2 at the upper bounding surface
     - self.sed_area: area of seafloor which is intercepted by this box.
     - self.area_fraction: area of seafloor which is intercepted by this relative to the total ocean floor area


    carbonate_system:
    ~~~~~~~~~~~~~~~~~

    If the reservoir group has a DIC and TA reservoir, and if the
    seawater_parameters key has been supplied as well, this keyword
    will add a carbonate_chemistry chemistry module to the reservoir
    group. The values of the carbonate system are available assign:

     - self.cs.H
     - self.cs.CA
     - self.cs.HCO3
     - self.cs.CO3
     - self.CO2aq

    seawater_parameters:
    ~~~~~~~~~~~~~~~~~~~~
                    
    If this optional parameter is specified, a SeaWaterConstants instance will
    be registered for this Reservoir as Reservoir.swc
    See the  SeaWaterConstants class for details how to specify the parameters, e.g.::
                    
        seawater_parameters = {"temperature": 2,
                           "pressure": 240,
                           "salinity" : 35,
                           },

    """

    def __init__(self, **kwargs) -> None:
        """Initialize a new reservoir group"""

        from esbmtk import (
            ExternalCode,
            Species,
            SeawaterConstants,
            Model,
            get_box_geometry_parameters,
            Q_,
        )

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "delta": ["None", (dict, str)],
            "concentration": ["None", (dict, str)],
            "mass": ["None", (str, dict)],
            "volume": ["None", (str, Q_)],
            "geometry": ["None", (str, list)],
            "plot": ["None", (str, dict)],
            "isotopes": [False, (dict, bool)],
            "seawater_parameters": ["None", (dict, str)],
            "carbonate_system": [False, (bool)],
            "register": ["None", (str, Model)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = [
            "name",
            "register",
            ["volume", "geometry"],
        ]

        if "concentration" in kwargs:
            self.species: list = list(kwargs["concentration"].keys())
        elif "mass" in kwargs:
            self.species: list = list(kwargs["mass"].keys())
        else:
            raise ReservoirGroupError("You must provide either mass or concentration")

        self.__initialize_keyword_variables__(kwargs)

        # legacy variable
        self.n = self.name
        self.mo = self.species[0].mo
        self.model = self.species[0].mo
        self.parent = self.register
        self.has_cs1 = False
        self.has_cs2 = False

        # geoemtry information
        if self.volume == "None":
            get_box_geometry_parameters(self)
        elif isinstance(self.volume, str):
            self.volume = Q_(self.volume).to(self.mo.v_unit)
        elif not isinstance(self.volume, Q_):
            raise ReservoirGroupError("Volume must be string or quantity")

        # register this group object in the global namespace
        # if self.mo.register == "local" and self.register == "None":
        #     self.register = self.mo
        self.__register_name_new__()

        # register a seawater_parameter instance if necessary
        if self.seawater_parameters != "None":
            if "temperature" in self.seawater_parameters:
                self.temperature = self.seawater_parameters["temperature"]
            else:
                self.temperature = 25
            if "salinity" in self.seawater_parameters:
                self.salinity = self.seawater_parameters["salinity"]
            else:
                self.salinity = 35
            if "pressure" in self.seawater_parameters:
                self.pressure = self.seawater_parameters["pressure"]
            else:
                self.pressure = 1

            SeawaterConstants(
                name="swc",
                temperature=self.temperature,
                pressure=self.pressure,
                salinity=self.salinity,
                register=self,
            )

        # dict with all default values
        self.cd: dict = {}
        for s in self.species:
            self.cd[s.name]: dict = {
                "mass": "None",
                "concentration": "None",
                "delta": "None",
                "plot": "yes",
                "isotopes": False,
            }

            # now we loop trough all keys for this reservoir and see
            # if we find a corresponding item in the kwargs
            for kcd, vcd in self.cd[s.name].items():  # kcd  = delta, plot, etc
                if kcd in self.kwargs and s in self.kwargs[kcd]:
                    # update the entry with the value provided in kwargs
                    # self.cd['SO4_name']['delta'] = self.kwargs['delta'][SO4]
                    self.cd[s.name][kcd] = self.kwargs[kcd][s]

        self.lor: list = []  # list of reservoirs in this group.
        # loop over all entries in species and create the respective reservoirs
        for s in self.species:
            if not isinstance(s, Species):
                raise ReservoirGroupError(f"{s.n} needs to be a valid species name")
            if s.flux_only:
                rtype = "flux_only"
            else:
                rtype = "regular"

            # create reservoir without registering it in the global name space
            a = Reservoir(
                name=f"{s.name}",
                register=self,
                species=s,
                delta=self.cd[s.n]["delta"],
                mass=self.cd[s.n]["mass"],
                concentration=self.cd[s.n]["concentration"],
                volume=self.volume,
                # geometry=self.geometry,
                plot=self.cd[s.n]["plot"],
                groupname=self.name,
                isotopes=self.cd[s.n]["isotopes"],
                rtype=rtype,
            )
            # register as part of this group
            self.lor.append(a)

        # depreceated
        if self.carbonate_system:
            # do some sanity checks:
            if not hasattr(self, "swc"):
                raise ReservoirGroupError(
                    f"{self.full_name} has no seawaterconstants instance"
                )
            if not hasattr(self, "DIC"):
                raise ReservoirGroupError(f"{self.full_name} has no DIC reservoir")

            if not hasattr(self, "TA"):
                raise ReservoirGroupError(f"{self.full_name} has no TA reservoir")

            ExternalCode(
                name="cs",
                species=Model.CO2,
                alias_list="H CA HCO3 CO3 CO2aq omega zsat".split(" "),
                vr_datafields=list(
                    [
                        self.swc.hplus,
                        self.swc.ca,
                        self.swc.hco3,
                        self.swc.co3,
                        self.swc.co2,
                        0.0,  # omega
                        0.0,  # zsat
                    ]
                ),
                function=calc_carbonates_2,
                function_input_data=list([self.DIC.c, self.TA.c]),
                function_params=list(
                    [
                        self.swc.K1,  # 0
                        self.swc.K2,  # 1
                        self.swc.KW,  # 2
                        self.swc.KB,  # 3
                        self.swc.boron,  # 4
                        self.swc.hplus,  # 5
                        self.swc.ca2,  # 6
                        self.swc.Ksp,  # 7
                        self.swc.Ksp0,  # 8
                        self.swc.zsat0,  # zsat0 after Boudreau 2010
                    ]
                ),
                register=self,
            )


class SourceSink(esbmtkBase):
    """
    This is a meta class to setup a Source/Sink objects. These are not
    actual reservoirs, but we stil need to have them as objects
    Example::

           Sink(name = "Pyrite",
               species = SO4,
               display_precision = number, optional, inherited from Model
               delta = number or str. optional defaults to "None"
           )

    where the first argument is a string, and the second is a reservoir handle

    """

    def __init__(self, **kwargs) -> None:
        from esbmtk import Species, Model, SourceSinkGroup

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "display_precision": [0.01, (int, float)],
            "register": ["None", (str, Model, SourceSinkGroup)],
            "delta": ["None", (int, float, str)],
            "isotopes": [False, (bool)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name", "species", "register"]

        self.__initialize_keyword_variables__(kwargs)

        self.loc: set[Connection] = set()  # set of connection objects

        # legacy names
        # if self.register != "None":
        #    self.full_name = f"{self.name}.{self.register.name}"

        self.parent = self.register
        self.n = self.name
        self.sp = self.species
        self.mo = self.species.mo
        self.u = self.species.mu + "/" + str(self.species.mo.bu)
        self.lio: list = []

        self.__register_name_new__()


class SourceSinkGroup(esbmtkBase):
    """
    This is a meta class to setup  Source/Sink Groups. These are not
    actual reservoirs, but we stil need to have them as objects
    Example::

           Sink(name = "Pyrite",
                species = [SO42, H2S],
                delta = {"SO4": 10}
                )

    where the first argument is a string, and the second is a reservoir handle
    """

    def __init__(self, **kwargs) -> None:
        from esbmtk import Model, Species, Source, Sink

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, list)],
            "delta": [{}, dict],
            "register": ["None", (str, Model)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name", "species", "register"]

        self.__initialize_keyword_variables__(kwargs)

        # legacy variables
        self.n = self.name
        self.parent = self.register
        self.loc: set[Connection] = set()  # set of connection objects

        # register this object in the global namespace
        self.mo = self.species[0].mo  # get model handle
        self.model = self.species[0].mo

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name_new__()

        self.lor: list = []  # list of sub reservoirs in this group

        # loop over species names and setup sub-objects
        for i, s in enumerate(self.species):
            if not isinstance(s, Species):
                raise SourceSinkGroupError(f"{s.n} needs to be a valid species name")

            delta = self.delta[s] if s in self.delta else "None"
            if type(self).__name__ == "SourceGroup":
                a = Source(
                    name=f"{s.name}",
                    register=self,
                    species=s,
                    delta=delta,
                )

            elif type(self).__name__ == "SinkGroup":
                a = Sink(
                    name=f"{s.name}",
                    register=self,
                    species=s,
                    delta=delta,
                )
            else:
                raise SourceSinkGroupError(
                    f"{type(self).__name__} is not a valid class type"
                )

            # register in local namespace
            self.lor.append(a)


class SinkGroup(SourceSinkGroup):
    """
    This is just a wrapper to setup a Sink object
    Example::

           Sink(name = "Pyrite",species =SO4)

    where the first argument is a string, and the second is a species handle
    """


class SourceGroup(SourceSinkGroup):
    """
    This is just a wrapper to setup a Source object
    Example::

           Sink(name = "SO4_diffusion", species ="SO4")

    where the first argument is a string, and the second is a species handle
    """


class Signal(esbmtkBase):
    """This class will create a signal which is
    described by its startime (relative to the model time), it's
    size (as mass) and duration, or as duration and
    magnitude. Furthermore, we can presribe the signal shape
    (square, pyramid) and whether the signal will repeat. You
    can also specify whether the event will affect the delta value.

    The data in the signal class will simply be added to the data in
    a given flux. So this class cannot be used for scaling (can we
    add this functionality?)

    Example::

          Signal(name = "Name",
                 species = Species handle,
                 start = "0 yrs",     # optional
                 duration = "0 yrs",  #
                 delta = 0,           # optional
                 stype = "addition"   # optional, currently the only type
                 shape = "square/pyramid/bell/filename"
                 mass/magnitude/filename  # give one
                 offset = '0 yrs',     #
                 scale = 1, optional,  #
                 offset = option #
                 reservoir = r-handle # optional, see below
                 source = s-handle optional, see below
                 display_precision = number, optional, inherited from Model
                 register,
                )

    Signals are cumulative, i.e., complex signals ar created by
    adding one signal to another (i.e., Snew = S1 + S2)

    The optional scaling argument will only affect the y-column data of
    external data files

    Signals are registered with a flux during flux creation,
    i.e., they are passed on the process list when calling the
    connector object.

    if the filename argument is used, you can provide a filename which
    contains the data to be used in csv format. The data will be
    interpolated to the model domain, and added to the already existing data.
    The external data need to be in the following format

      Time, Rate, delta value
      0,     10,   12

      i.e., the first row needs to be a header line

    All time data in the csv file will be treated as realative time
    (i.e., the start time will be mapped to zero). Use the offset
    keyword to shift the external signal data in the time domain.

    Last but not least, you can provide an optional reservoir name. In
    this case, the signal will create a source as (signal_name_source)
    and the connection to the specified reservoir. If you build a
    complex signal do this as the last step. If you additionally
    provide a source name the connection will be made between the
    provided source (this can be useful if you use source groups).


    This class has the following methods

      Signal.repeat()
      Signal.plot()
      Signal.info()

    """

    def __init__(self, **kwargs) -> None:
        """Parse and initialize variables"""

        from esbmtk import Q_, Species, Source, Sink, Reservoir, Model

        # provide a list of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "start": ["0 yrs", (str, Q_)],
            "duration": ["1 yr", (str, Q_)],
            "species": ["None", (Species)],
            "delta": [0, (int, float)],
            "stype": [
                "addition",
                (str),
            ],
            "shape": ["None", (str)],
            "filename": ["None", (str)],
            "mass": ["None", (str, Q_)],
            "magnitude": ["None", (str, Q_)],
            "offset": ["0 yrs", (str, Q_)],
            "plot": ["no", (str)],
            "scale": [1, (int, float)],
            "display_precision": [0.01, (int, float)],
            "reservoir": ["None", (Source, Sink, Reservoir, str)],
            "source": ["None", (Source, Sink, Reservoir, str)],
            "legend_right": ["None", (str)],
            "register": ["None", (str, Model)],
            "isotopes": [False, (bool)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = [
            "name",
            ["duration", "filename"],
            "species",
            ["shape", "filename"],
            ["magnitude", "mass", "filename"],
            "register",
        ]

        self.__initialize_keyword_variables__(kwargs)

        # list of signals we are based on
        self.los: list[Signal] = []

        # convert units to model units
        self.st: tp.Union[int, float] = int(
            Q_(self.start).to(self.species.mo.t_unit).magnitude
        )  # start time

        if "mass" in self.kwargs:
            self.mass = Q_(self.mass).to(self.species.mo.m_unit).magnitude
        elif "magnitude" in self.kwargs:
            self.magnitude = Q_(self.magnitude).to(self.species.mo.f_unit).magnitude

        self.duration = int(Q_(self.duration).to(self.species.mo.t_unit).magnitude)

        self.offset = Q_(self.offset).to(self.species.mo.t_unit).magnitude

        if self.duration / self.species.mo.max_step < 10:
            print("\n\n ------------------------------------------------\n")
            print("\n\n   W A R N I N G \n\n")
            print("Your signal duration is covered by less than 10")
            print("Intergration steps. This may not be what you want")
            print("Consider adjusting the max_step parameter of the model object\n\n")
            print("\n\n ------------------------------------------------\n")

        # legacy name definitions
        self.full_name = ""
        self.l: int = self.duration
        self.n: str = self.name  # the name of the this signal
        self.sp: Species = self.species  # the species
        self.mo: Model = self.species.mo  # the model handle
        self.model = self.mo
        self.parent = self.register
        self.ty: str = self.stype  # type of signal
        self.sh: str = self.shape  # shape the event
        self.d: float = self.delta  # delta value offset during the event
        self.kwd: dict[str, any] = self.kwargs  # list of keywords
        self.led: list = []

        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        self.data = self.__init_signal_data__()
        self.m = self.data.m
        if self.isotopes:
            self.l = self.data.l

        self.data.n: str = self.name + "_data"  # update the name of the signal data
        self.legend_left = self.data.legend_left
        self.legend_right = self.data.legend_right
        # update isotope values
        # self.data.li = get_l_mass(self.data.m, self.data.d, self.sp.r)
        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo
        self.__register_name_new__()
        self.mo.los.append(self)  # register with model

        # in case we deal with a sink or source signal
        if self.reservoir != "None":
            self.__apply_signal__()

    def __init_signal_data__(self) -> None:
        """1. Create a vector which contains the signal data. The vector length
           can exceed the modelling domain.
        2. Trim the signal vector in such a way that it fits within the
           modelling domain
        3. Create an empty flux and replace it with the signal vector data.

        Note that this flux will then be added to an existing flux.

        """

        from esbmtk import Flux

        # these are signal times, not model time
        self.length: int = int(round(self.duration / self.mo.dt))

        # create signal vector
        if self.sh == "square":
            self.__square__(0, self.length)

        elif self.sh == "pyramid":
            self.__pyramid__(0, self.length)
        elif self.sh == "bell":
            self.__bell__(0, self.length)
        elif "filename" in self.kwargs:  # use an external data set
            self.length = self.__int_ext_data__()
        else:
            raise SignalError(
                "argument needs to be either square/pyramid, or an ExternalData object. "
            )

        # create a dummy flux we can act up
        self.nf: Flux = Flux(
            name=self.n + "_data",
            species=self.sp,
            rate=f"0 {self.sp.mo.f_unit}",
            delta=0,
            isotopes=self.isotopes,
            save_flux_data=True,
            register=self,
        )

        # remove signal fluxes from global flux list
        self.mo.lof.remove(self.nf)

        # map into model space
        insert_start_time = self.st - self.mo.offset
        insert_stop_time = insert_start_time + self.duration

        dt1 = int((self.st - self.mo.offset - self.mo.start))
        dt2 = int((self.st + self.duration - self.mo.stop - self.mo.offset))

        # This fails if actual model steps != dt
        model_start_index = int(max(insert_start_time / self.mo.dt, 0))
        model_stop_index = int(min((self.mo.steps + dt2 / self.mo.dt), self.mo.steps))
        signal_start_index = int(min(dt1, 0) * -1)
        signal_stop_index = int(self.length - max(0, dt2))

        if self.mo.debug:
            print(
                f"dt1 = {dt1}, dt2 = {dt2}, offset = {self.mo.offset}"
                f"insert start time = {insert_start_time} "
                f"insert_stop time = {insert_stop_time} "
                f"duration = {self.duration}\n"
                f"msi = {model_start_index}, msp = {model_stop_index} "
                f"model num_steps = {model_stop_index-model_start_index}\n"
                f"ssi = {signal_start_index}, ssp = {signal_stop_index} "
                f"signal num_steps = {signal_stop_index-signal_start_index}\n"
            )

        if signal_start_index < signal_stop_index:
            self.nf.m[model_start_index:model_stop_index] = self.s_m[
                signal_start_index:signal_stop_index
            ]
            if self.nf.isotopes:
                self.nf.l[model_start_index:model_stop_index] = self.s_l[
                    signal_start_index:signal_stop_index
                ]
        return self.nf

    def __square__(self, s, e) -> None:
        """Create Square Signal"""

        self.s_m: NDArrayFloat = np.zeros(e - s)
        self.s_d: NDArrayFloat = np.zeros(e - s)

        if "mass" in self.kwd:
            h = self.mass / self.duration  # get the height of the square
            self.magnitude = h

        elif "magnitude" in self.kwd:
            h = self.magnitude
            self.mass = h * self.duration
        else:
            raise SignalError("You must specify mass or magnitude of the signal")

        self.s_m = self.s_m + h  # add this to the section
        self.s_d = self.s_d + self.d  # add the delta offset
        self.s_l = get_l_mass(self.s_m, self.s_d, self.sp.r)

    def __pyramid__(self, s, e) -> None:
        """Create pyramid type Signal

        s = start index
        e = end index
        """

        if "mass" in self.kwd:
            h = 2 * self.mass / self.duration  # get the height of the pyramid

        elif "magnitude" in self.kwd:
            h = self.magnitude
        else:
            raise SignalError("You must specify mass or magnitude of the signal")

        # create pyramid
        c: int = int(round((e - s) / 2))  # get the center index for the peak
        x: NDArrayFloat = np.array([0, c, e - s])  # setup the x coordinates
        y: NDArrayFloat = np.array([0, h, 0])  # setup the y coordinates

        d: NDArrayFloat = np.array([0, self.d, 0])  # setup the d coordinates
        xi = np.arange(0, e - s)  # setup the points at which to interpolate

        self.s_m: NDArrayFloat = np.interp(xi, x, y)  # interpolate flux
        self.s_d: NDArrayFloat = np.interp(xi, x, d)  # interpolate delta
        self.s_l = get_l_mass(self.s_m, self.s_d, self.sp.r)

    def __bell__(self, s, e) -> None:
        """Create a bell curve type signal

        s = start index
        e = end index

        Note that the area under the curve equals one.
        So we can scale the result simply with mass
        """

        import sys

        c: int = int(round((e - s) / 2))  # get the center index for the peak
        x: NDArrayFloat = np.arange(-c, c + 1, 1)
        e: float = math.e
        pi: float = math.pi
        mu: float = 0
        phi: float = c / 4

        print(f"mu = {mu} ,phi = {phi}")
        print(f"x[0] = {x[0]}, x[-1] = {x[-1]}")
        print(sys.float_info)

        a = -((x - mu) ** 2) / (2 * phi**2)
        # get bell curve
        self.s_m = 1 / (phi * math.sqrt(2 * pi)) * e**a
        self.s_d = self.s_m * self.delta / max(self.s_m)
        self.s_l = self.s_m

        if "mass" in self.kwargs:
            self.s_m = self.s_m * self.mass
        elif "magnitude" in self.kwargs:
            self.s_m = self.s_m * self.magnitude / max(self.s_m)
        else:
            raise SignalError("Bell type signal require either mass or magnitude")

    def __int_ext_data__(self) -> None:
        """Interpolate External data as a signal. Unlike the other signals,
        this will replace the values in the flux with those read from the
        external data source. The external data need to be in the following format

        Time [units], Rate [units], delta value [units]
        0,     10,   12

        i.e., the first row needs to be a header line

        """

        from . import Q_

        if not os.path.exists(self.filename):  # check if the file is actually there
            raise SignalError(f"Cannot find file {self.filename}")

        # read external dataset
        df = pd.read_csv(self.filename)
        # print(f"df = {df.head()}")

        # get unit information from each header
        xh = df.columns[0].split("[")[1].split("]")[0]
        yh = df.columns[1].split("[")[1].split("]")[0]
        zh = df.columns[2].split("[")[1].split("]")[0] if len(df.columns) > 2 else None
        # create the associated quantities
        xq = Q_(xh)
        yq = Q_(yh)

        # add these to the data we are are reading
        self.s_time: NDArrayFloat = df.iloc[:, 0].to_numpy() * xq
        self.s_data: NDArrayFloat = df.iloc[:, 1].to_numpy() * yq

        if zh:
            # delta is assumed to be without units
            self.s_delta: NDArrayFloat = df.iloc[:, 2].to_numpy()

        # map into model units, and strip unit information
        self.s_time = self.s_time.to(self.mo.t_unit).magnitude
        # self.s_data = self.s_data.to(self.mo.f_unit).magnitude * self.scale

        if isinstance(yq, Q_):
            # test what type of Quantity we have
            if yq.check(["dimensionless"]):  # dimensionless
                self.s_data = self.s_data.magnitude
            elif yq.check(["volume]/[time"]):  # flux
                self.s_data = self.s_data.to(self.mo.r_unit).magnitude
            elif yq.check(["mass] / [time"]):  # flux
                self.s_data = self.s_data.to(self.mo.f_unit).magnitude
            elif yq.check(["[mass]/[volume]"]):  # concentration
                self.s_data = self.s_data.to(self.mo.c_unit).magnitude
            else:
                SignalError(f"No conversion to model units for {self.scale} specified")

        self.s_data = self.s_data * self.scale

        self.st: float = self.s_time[0]  # start time
        self.et: float = self.s_time[-1]  # end time
        self.duration = int(round((self.et - self.st)))
        num_steps = int(self.duration / self.mo.dt)
        # setup the points at which to interpolate
        xi = np.linspace(self.st, self.et, num_steps)

        self.s_m: NDArrayFloat = np.interp(
            xi, self.s_time, self.s_data
        )  # interpolate flux
        if zh:
            self.s_d: NDArrayFloat = np.interp(xi, self.s_time, self.s_delta)
            self.s_l = get_l_mass(self.s_m, self.s_d, self.sp.r)
        else:
            self.s_l: NDArrayFloat = np.zeros(num_steps)

        return num_steps

    def __apply_signal__(self) -> None:
        """In case we deal with a source  signal, we need
        to create a source, and connect signal, source and reservoir
        Maybe this logic should be me moved elsewhere?
        """

        from esbmtk import Source, Connect

        if self.source == "None":
            self.source = Source(name=f"{self.name}_Source", species=self.sp)

        Connect(
            source=self.source,  # source of flux
            sink=self.reservoir,  # target of flux
            rate="0 mol/yr",  # flux rate
            signal=self,  # list of processes
            plot="no",
        )

    def __add__(self, other):
        """allow the addition of two signals and return a new signal"""

        ns = cp.deepcopy(self)

        # add the data of both fluxes
        # get delta of self
        sd = get_delta(self.data.l, self.data.m - self.data.l, self.data.sp.r)
        od = get_delta(other.data.l, other.data.m - other.data.l, other.data.sp.r)
        ns.data.m = self.data.m + other.data.m
        ns.data.l = get_l_mass(ns.data.m, sd + od, ns.data.sp.r)

        ns.n: str = self.n + "_and_" + other.n
        # print(f"adding {self.n} to {other.n}, returning {ns.n}")
        ns.data.n: str = self.n + "_and_" + other.n + "_data"
        ns.st = min(self.st, other.st)
        ns.l = max(self.l, other.l)
        ns.sh = "compound"
        ns.los.append(self)
        ns.los.append(other)

        return ns

    def repeat(self, start, stop, offset, times) -> None:
        """This method creates a new signal by repeating an existing signal.
        Example::

            new_signal = signal.repeat(start,   # start time of signal slice to be repeated
                                       stop,    # end time of signal slice to be repeated
                                       offset,  # offset between repetitions
                                       times,   # number of time to repeat the slice
                                       )

        """

        ns: Signal = cp.deepcopy(self)
        ns.n: str = self.n + f"_repeated_{times}_times"
        ns.data.n: str = self.n + f"_repeated_{times}_times_data"
        start: int = int(start / self.mo.dt)  # convert from time to index
        stop: int = int(stop / self.mo.dt)
        offset: int = int(offset / self.mo.dt)
        ns.start: float = start
        ns.stop: float = stop
        ns.offset: float = stop - start + offset
        ns.times: float = times
        ns.ms: NDArrayFloat = self.data.m[start:stop]  # get the data slice we are using
        ns.ds: NDArrayFloat = self.data.d[start:stop]

        diff = 0
        for _ in range(times):
            start += ns.offset
            stop += ns.offset
            if start > len(self.data.m):
                break
            elif stop > len(self.data.m):  # end index larger than data size
                diff: int = stop - len(self.data.m)  # difference
                stop -= diff
                lds: int = len(ns.ds) - diff
            else:
                lds: int = len(ns.ds)

            ns.data.m[start:stop]: NDArrayFloat = ns.data.m[start:stop] + ns.ms[:lds]
            ns.data.d[start:stop]: NDArrayFloat = ns.data.d[start:stop] + ns.ds[:lds]

        # and recalculate li and hi
        ns.data.l: NDArrayFloat
        ns.data.h: NDArrayFloat
        [ns.data.l, ns.data.h] = get_imass(ns.data.m, ns.data.d, ns.data.sp.r)
        return ns

    def __register_with_flux__(self, flux) -> None:
        """Register this signal with a flux. This should probably be done
        through a process!

        """

        from esbmtk import Flux, Species

        self.fo: Flux = flux  # the flux handle
        self.sp: Species = flux.sp  # the species handle
        # list of processes
        flux.lop.append(self)

    def __call__(self, t) -> list[float, float]:
        """Return Signal value at time t (mass and mass for light
        isotope). This will work as long a t is a multiple of dt, and i = t.
        may extend this by addding linear interpolation but that will
        be costly

        """

        import numpy as np

        m = np.interp(t, self.mo.time, self.data.m)
        if self.isotopes:
            l = np.interp(t, self.mo.time, self.data.l)
        else:
            l = 0

        return [m, l]

    def __plot__(self, M: Model, ax) -> None:
        """Plot instructions.
        M: Model
        ax: matplotlib axes handle
        """

        from esbmtk import set_y_limits

        # convert time and data to display units
        x = (M.time * M.t_unit).to(M.d_unit).magnitude
        y1 = (self.data.m * M.f_unit).to(self.data.plt_units).magnitude
        legend = f"{self.legend_left} [{M.f_unit}]"

        # # test for plt_transform
        # if self.plot_transform_c != "None":
        #     if Callable(self.plot_transform_c):
        #         y1 = self.plot_transform_c(self.c)
        #     else:
        #         raise ValueError("Plot transform must be a function")

        # plot first axis
        ln1 = ax.plot(x[1:-2], y1[1:-2], color="C0", label=self.legend_left)
        ax.set_xlabel(f"{M.time_label} [{M.d_unit:~P}]")
        ax.set_ylabel(legend)
        ax.spines["top"].set_visible(False)
        handler1, label1 = ax.get_legend_handles_labels()

        # plot second axis
        if self.isotopes:
            axt = ax.twinx()
            y2 = self.d  # no conversion for isotopes
            ln2 = axt.plot(x[1:-2], y2[1:-2], color="C1", label=self.legend_right)
            axt.set_ylabel(self.data.ld)
            set_y_limits(axt, M)
            x.spines["top"].set_visible(False)
            # set combined legend
            handler2, label2 = axt.get_legend_handles_labels()
            legend = axt.legend(handler1 + handler2, label1 + label2, loc=0).set_zorder(
                6
            )
        else:
            ax.legend()
            ax.spines["right"].set_visible(False)
            ax.yaxis.set_ticks_position("left")
            ax.xaxis.set_ticks_position("bottom")


class DataField(esbmtkBase):
    """
    DataField: Datafields can be used to plot data which is computed after
    the model finishes in the overview plot windows. Therefore, datafields will
    plot in the same window as the reservoir they are associated with.
    Datafields must share the same x-axis is the model, and can have up to two
    y axis.

    Example::

             DataField(name = "Name"
                       register = Model handle,
                       y1_data = NDArrayFloat or list of arrays
                       y1_label = Data label(s)
                       y1_legend = Y-Axis Label
                       y1_type = "plot", | "scatter"
                       y2_data = NDArrayFloat    # optional
                       y2_legend = Y-Axis label # optional
                       y2_label = Data legend(s) # optional
                       y2_type = "plot", | "scatter"
                       common_y_scale = "no",  #optional, default "no"
                       display_precision = number, optional, inherited from Model
                       )

    Note that Datafield data is not mapped to model units. Care must be taken
    that the data units match the model units.

    The instance provides the following data

    Name.x    = X-axis = model X-axis
    Name.y1_data
    Name.y1_label
    Name.y1_legend

    Similarly for y2
    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """Initialize this instance"""

        from . import Reservoir_no_set, VirtualReservoir, ExternalCode, Model

        # dict of all known keywords and their type
        self.defaults: dict[str, list(str, tuple)] = {
            "name": ["None", (str)],
            "register": [
                "None",
                (
                    Model,
                    Reservoir,
                    ReservoirGroup,
                    Reservoir_no_set,
                    VirtualReservoir,
                    ExternalCode,
                ),
            ],
            "associated_with": [
                "None",
                (
                    Model,
                    Reservoir,
                    ReservoirGroup,
                    Reservoir_no_set,
                    VirtualReservoir,
                    ExternalCode,
                ),
            ],
            "y1_data": ["None", (np.ndarray, list)],
            "x1_data": ["None", (np.ndarray, list, str)],
            "x1_as_time": [False, (bool)],
            "y1_label": ["Not Provided", (str, list)],
            "y1_legend": ["Not Provided", (str)],
            "y1_type": ["plot", (str, list)],
            "y2_data": ["None", (str, np.ndarray, list)],
            "x2_data": ["None", (np.ndarray, list, str)],
            "x2_as_time": [False, (bool)],
            "y2_label": ["Not Provided", (str, list)],
            "y2_legend": ["Not Provided", (str)],
            "y2_type": ["plot", (str, list)],
            "common_y_scale": ["no", (str)],
            "display_precision": [0.01, (int, float)],
            "title": ["None", (str)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["name", ["register", "associated_with"], "y1_data"]

        # provide a dictionary entry for a keyword specific error message
        # see esbmtkBase.__initerrormessages__()
        self.__initialize_keyword_variables__(kwargs)

        if self.register == "None":
            self.register = self.associated_with

        # set legacy variables
        self.legend_left = self.y1_legend
        self.isotopes = False
        self.parent = self.register

        # if self.associated_with == "None":
        #     self.associated_with = self.mo.lor[0]

        # self.mo = self.associated_with.mo
        self.mo = self.register.mo
        self.model = self.mo
        if self.associated_with == "None":
            if isinstance(self.register, Model):
                self.associated_with = self.register.lor[0]
            elif isinstance(self.register, Reservoir):
                self.associated_with = self.register
            else:
                raise DataFieldError(
                    "Set associated_with or register to a reservoir name"
                )

        if isinstance(self.associated_with, Reservoir):
            self.plt_units = self.associated_with.plt_units
        elif isinstance(self.associated_with, ReservoirGroup):
            self.plt_units = self.associated_with.lor[0].plt_units
        else:
            raise ValueError("This needs fixing")

        # if not isinstance(self.y2_data, str):
        #     self.d = self.y2_data
        #     self.legend_right = self.y2_legend
        #     self.ld = self.y2_label

        self.n = self.name
        self.led = []

        # register with reservoir
        # self.associated_with.ldf.append(self)
        # register with model. needed for print_reservoirs
        self.mo.ldf.append(self)
        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        self.__register_name_new__()
        if self.mo.state == 0:
            print("")
            print(
                "---------------------------------------------------------------------------\n\n"
            )
            print(
                "Warning, you are initializing a datafield before the model results are known\n\n"
            )
            print(
                "---------------------------------------------------------------------------"
            )

    def __write_data__(
        self,
        prefix: str,
        start: int,
        stop: int,
        stride: int,
        append: bool,
        directory: str,
    ) -> None:
        """To be called by write_data and save_state"""

        from pathlib import Path

        p = Path(directory)
        p.mkdir(parents=True, exist_ok=True)

        # some short hands
        mo = self.mo  # model handle

        mtu = f"{mo.t_unit:~P}"
        rn = self.n  # reservoir name
        mn = self.mo.n  # model name

        if self.sp.mo.register == "None":
            fn = f"{directory}/{prefix}{mn}_{rn}.csv"  # file name
        elif self.sp.mo.register == "local":
            fn = f"{directory}/{prefix}{rn}.csv"  # file name
        else:
            raise DataFieldError(
                f"Model register keyword must be 'None'/'local' not {self.sp.mo.register}"
            )

        # build the dataframe
        df: pd.dataframe = DataFrame()

        df[f"{self.n} Time [{mtu}]"] = self.mo.time[start:stop:stride]  # time
        df[f"{self.n} {self.y1_legend}"] = self.y1_data[start:stop:stride]  # y1 data

        if self.y2_data != "None":
            df[f"{self.n} {self.y2_legend}"] = self.y2_data[
                start:stop:stride
            ]  # y2_data

        file_path = Path(fn)
        if append and file_path.exists():
            df.to_csv(file_path, header=False, mode="a", index=False)
        else:
            df.to_csv(file_path, header=True, mode="w", index=False)
        return df

    def __unify_data__(self, M: Model, x, y, label) -> tuple(list, list):
        """The input data for the DataField Class can be either missing, be a
        single array, or a list of arrays.  The module analyzes the data and
        modifies is in such a way that it can be used by the __plot__ method
        without further adjustments.

        :param x: str, array, list, withe the x-data
        :param y: str, array, list, withe the y-data

        :return x,y: as list
        """
        import numpy as np

        # print(f"ud0: type x = {type(x)}, len = {len(x)}")
        # print(f"ud1: type y = {type(x)}, len = {len(y)}")
        # consider the y-axis first
        if isinstance(y, str):
            ValueError("Provide at least one data-set")
        elif isinstance(y, np.ndarray):
            y = [y]
            y_l = 1
        elif isinstance(y, list):
            y_l = len(y)
        else:
            raise DataFieldError("Y data needs to be array, numpy array or list")

        # consider the x-axis next
        if isinstance(x, str):  # no x-data has been provided
            if y_l == 1:  # single y data
                x = [M.time]
                # print(f" no x-data y_l = {y_l}")
            else:  # mutiple y data
                # print(f" no x-data mutiple y-data y_l = {y_l}")
                x = []
                for e in range(y_l):
                    x.append(M.time)

        elif isinstance(x, np.ndarray):
            # print(f"np_array, y_l = {y_l}")
            xx = []
            if y_l == 1:  # single y data
                xx.append(x)
            else:  # mutiple y data
                # print(f" no x-data mutiple y-data y_l = {y_l}")
                for e in range(y_l):
                    xx.append(x)
            x = xx
            # print(f"after: {type(x)}")
        elif isinstance(x, list):  # assume that lists match
            if len(x) != len(y):
                raise DataFieldError(f"Y data needs to match x data for {label}")
        else:
            raise DataFieldError(
                f"Y data needs to be array, numpy array or list for {label}"
            )

        if y_l == 1:
            if not isinstance(label, list):
                label = [label]
        elif y_l > 1:
            if not isinstance(label, list):
                ll = []
                for e in y_l:
                    ll.append(label)
                label = ll

        return x, y, label

    def __plot_data__(self, ax, x, y, t, l, i) -> None:
        """Plot data either as line of scatterplot

        :param ax: axis handle
        :param x: x data
        :param y: y data
        :param t: plot type
        :param l: label str
        :param i: index

        """

        if t == "plot":
            ax.plot(x, y, color=f"C{i}", label=l)
        else:
            ax.scatter(x, y, color=f"C{i}", label=l)

    def __plot__(self, M: Model, ax) -> None:
        """Plot instructions.
        M: Model
        ax: matplotlib axes handle
        """

        # plot external data first
        for i, d in enumerate(self.led):
            time = (d.x * M.t_unit).to(M.d_unit).magnitude
            # yd = (d.y * M.c_unit).to(self.plt_units).magnitude
            leg = f"{d.legend}"
            ax.scatter(time[1:-2], d.y[1:-2], color=f"C{i}", label=leg)

        self.x1_data, self.y1_data, self.y1_label = self.__unify_data__(
            M,
            self.x1_data,
            self.y1_data,
            self.y1_label,
        )

        ymin = list()
        ymax = list()
        for i, d in enumerate(self.y1_data):  # loop over datafield list
            if self.x1_as_time:
                x1 = (self.x1_data[i] * M.t_unit).to(M.d_unit).magnitude
            else:
                x1 = self.x1_data[i]

            self.__plot_data__(
                ax,
                x1,
                self.y1_data[i],
                self.y1_type,
                self.y1_label[i],
                i,
            )
            last_i = i
            u, v = ax.get_ylim()
            ymin.append(u)
            ymax.append(v)
            # set_y_limits(ax, self)
        # add any external data if present
        ymin = min(ymin)
        ymax = max(ymax)
        ax.set_ylim([ymin, ymax])

        last_i = i
        ax.set_xlabel(f"{M.time_label} [{M.d_unit:~P}]")
        ax.set_ylabel(self.y1_legend)
        # remove unnecessary frame species
        ax.spines["top"].set_visible(False)
        handler1, label1 = ax.get_legend_handles_labels()

        # test if independeny y_data is present
        ymin = list()
        ymax = list()
        if not isinstance(self.y2_data, str):
            self.x2_data, self.y2_data, self.y2_label = self.__unify_data__(
                M,
                self.x2_data,
                self.y2_data,
                self.y2_label,
            )
            axt = ax.twinx()
            for i, d in enumerate(self.y2_data):  # loop over datafield list
                if self.x2_as_time:
                    x2 = (self.x1_data[i] * M.t_unit).to(M.d_unit).magnitude
                else:
                    x2 = self.x1_data[i]
                self.__plot_data__(
                    axt,
                    x2,
                    self.y2_data[i],
                    self.y2_type,
                    self.y2_label[i],
                    i + last_i + 1,
                )
            u, v = axt.get_ylim()
            ymin.append(u)
            ymax.append(v)
            ymin = min(ymin)
            ymax = max(ymax)
            axt.set_ylim([ymin, ymax])

            axt.set_xlabel(f"{M.time_label} [{M.d_unit:~P}]")
            axt.set_ylabel(self.y2_legend)
            # remove unnecessary frame species
            axt.spines["top"].set_visible(False)
            handler2, label2 = axt.get_legend_handles_labels()
            legend = axt.legend(handler1 + handler2, label1 + label2, loc=0).set_zorder(
                6
            )
            axt.legend()
        else:
            ax.legend(handler1, label1)
            ax.spines["right"].set_visible(False)
            ax.yaxis.set_ticks_position("left")
            ax.xaxis.set_ticks_position("bottom")
        if self.title != "None":
            ax.set_title(self.title)
        else:
            ax.set_title(self.register.full_name)


class Reservoir_no_set(ReservoirBase):
    """This class is similar to a regular reservoir, but we make no
    assumptions about the type of data contained. I.e., all data will be
    left alone

    """

    def __init__(self, **kwargs) -> None:
        """The original class will calculate delta and concentration from mass
        an d and h and l. Since we want to use this class without a
        priory knowledge of how the reservoir arrays are being used we
        overwrite the data generated during initialization with the
        values provided in the keywords

        """

        from esbmtk import (
            ConnectionGroup,
            Species,
            SourceGroup,
            SinkGroup,
            ReservoirGroup,
        )

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "plot_transform_c": ["None", (str, Callable)],
            "legend_left": ["None", (str)],
            "plot": ["yes", (str)],
            "groupname": ["None", (str)],
            "function": ["None", (str, Callable)],
            "display_precision": [0.01, (int, float)],
            "register": [
                "None",
                (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, str),
            ],
            "full_name": ["None", (str)],
            "isotopes": [False, (bool)],
            "volume": ["None", (str, int, float)],
            "vr_datafields": [{}, (dict)],
            "function_input_data": (list, str),
            "function_params": [list(), (list, str)],
            "geometry": ["None", (list, str)],
            "alias_list": ["None", (list, str)],
            "ref_flux": ["None", (list, str)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = [
            "name",
            "species",
        ]

        self.__validateandregister__(kwargs)
        self.__set_legacy_names__(kwargs)

        self.isotopes = False
        self.mu: str = self.sp.e.mass_unit  # massunit xxxx
        self.plt_units = self.mo.c_unit
        # save the unit which was provided by the user for display purposes

        # ------------------------------------------

        # left y-axis label
        self.lm: str = f"{self.species.n} [{self.mu}/l]"

        self.mo.lor.append(self)  # add this reservoir to the model
        # self.mo.lor.remove(self)
        # but lets keep track of  virtual reservoir in lvr.

        # this should be done in aux-inits of a derived class if necessary
        # self.mo.lvr.append(self)
        # print(f"added {self.name} to lvr 1")
        # register instance name in global name space
        self.__register_name_new__()

        self.__aux_inits__()
        self.state = 0


class ExternalCode(Reservoir_no_set):
    """This class can be used to implement user provided functions. The
    data inside a VR_no_set instance will only change in response to a
    user provided function but will otherwise remain unaffected. That is,
    it is up to the user provided function to manage changes in reponse to
    external fluxes. A VR_no_set is declared in the following way::

        ExternalCode(
                    name="cs",     # instance name
                    species=CO2,   # species, must be given
                    # the vr_data_fields contains any data that is referenced inside the
                    # function, rather than passed as argument, and all data that is
                    # explicitly referenced by the model
                    vr_datafields :dict ={"Hplus": self.swc.hplus,
                                          "Beta": 0.0},
                    function=calc_carbonates, # function reference, see below
                    fname = function name as string
                    function_input_data="DIC TA",
                    # Note that parameters must be individual float values
                    function_params:tuple(float)
                    # list of return values
                    return_values={  # these must be known speces definitions
                                  "Hplus": rg.swc.hplus,
                                   "zsnow": float(abs(kwargs["zsnow"])),
                                   },
                    register=rh # reservoir_handle to register with.
                )

        the dict keys of vr_datafields will be used to create alias
        names which can be used to access the respective variable


    The general template for a user defined function is a follows::

      def calc_carbonates(i: int, input_data: list, vr_data: list, params: list) -> None:
          # i = index of current timestep
          # input_data = list of np.arrays, typically data from other Reservoirs
          # vr_data = list of np.arrays created during instance creation (i.e. the vr data)
          # params = list of float values (at least one!)
          pass
       return

    Note that this function should not return any values, and that all input fields must have
    at least one entry!
                                   
    """

    def __init__(self, **kwargs) -> None:
        """The original class will calculate delta and concentration from mass
        an d and h and l. Since we want to use this class without a
        priory knowledge of how the reservoir arrays are being used we
        overwrite the data generated during initialization with the
        values provided in the keywords

        """

        from esbmtk import (
            ConnectionGroup,
            Species,
            SourceGroup,
            SinkGroup,
            ReservoirGroup,
            Model,
        )
        from typing import Callable

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "plot_transform_c": ["None", (str, Callable)],
            "legend_left": ["None", (str)],
            "plot": ["yes", (str)],
            "groupname": ["None", (str)],
            "function": ["None", (Callable, str)],
            "display_precision": [0.01, (int, float)],
            "register": [
                "None",
                (
                    SourceGroup,
                    SinkGroup,
                    ReservoirGroup,
                    ConnectionGroup,
                    GasReservoir,
                    Model,
                    str,
                ),
            ],
            "full_name": ["None", (str)],
            "isotopes": [False, (bool)],
            "volume": ["None", (str, int, float)],
            "vr_datafields": ["None", (dict, str)],
            "function_input_data": ["None", (str, list)],
            "function_params": ["None", (tuple)],
            "fname": ["None", (str)],
            "geometry": ["None", (list, str)],
            "alias_list": ["None", (list, str)],
            "ftype": ["None", (str)],
            "ref_flux": ["None", (list, str)],
            "return_values": ["None", (list)],
            "arguments": ["None", (str, list)],
            "r_s": ["None", (str, ReservoirGroup)],
            "r_d": ["None", (str, ReservoirGroup)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = [
            "name",
            "species",
            "register",
        ]

        self.__initialize_keyword_variables__(kwargs)

        self.__set_legacy_names__(kwargs)
        self.lro: list = []  # list of all return objects.
        self.mu: str = self.sp.e.mass_unit  # massunit xxxx
        self.plt_units = self.mo.c_unit

        # left y-axis label
        self.lm: str = f"{self.species.n} [{self.mu}/l]"
        self.mo.lor.append(self)  # add this reservoir to the model
        self.__register_name_new__()
        self.state = 0
        name = f"{self.full_name}_generic_function".replace(".", "_")
        logging.info(f"creating {name}")

        if self.vr_datafields != "None":
            self.alias_list = list(self.vr_datafields.keys())
            # initialize data fields
            self.vr_data = list()
            for e in self.vr_datafields.values():
                if isinstance(e, (float, int)):
                    self.vr_data.append(np.full(self.mo.steps, e, dtype=float))
                else:
                    self.vr_data.append(e)

        self.mo.lpc_r.append(self)
        # self.mo.lpc_r.append(self.gfh)

        self.mo.lor.remove(self)
        # but lets keep track of  virtual reservoir in lvr.
        self.mo.lvr.append(self)

        # print(f"added {self.name} to lvr 2")

        # create temporary memory if we use multiple solver iterations
        if self.mo.number_of_solving_iterations > 0:
            for i, d in enumerate(self.vr_data):
                setattr(self, f"vrd_{i}", np.empty(0))

        if self.alias_list != "None":
            self.create_alialises()

    def create_alialises(self) -> None:
        """Register  alialises for each vr_datafield"""

        for i, a in enumerate(self.alias_list):
            # print(f"{a} = {self.vr_data[i][0]}")
            setattr(self, a, self.vr_data[i])

    def append(self, **kwargs) -> None:
        """This method allows to update GenericFunction parameters after the
        VirtualReservoir has been initialized. This is most useful
        when parameters have to reference other virtual reservoirs
        which do not yet exist, e.g., when two virtual reservoirs have
        a circular reference.

        Example::

             VR.update(a1=new_parameter, a2=new_parameter)

        """

        allowed_keys: list = ["function_input_data, function_params"]
        # loop over provided kwargs
        for key, value in kwargs.items():
            if key not in allowed_keys:
                raise ESBMTKFunctionError(
                    "you can only change function_input_data, or function_params"
                )
            else:
                getattr(self, key).append(value)

    def __reset_state__(self):
        """Copy the last value to the first position so that we can restart the computation"""

        for i, d in enumerate(self.vr_data):
            d[0] = d[-2]
            setattr(
                self,
                f"vrd_{i}",
                np.append(getattr(self, f"vrd_{i}"), d[0 : -2 : self.mo.reset_stride]),
            )

    def __merge_temp_results__(self) -> None:
        """Once all iterations are done, replace the data fields
        with the saved values

        """

        # print(f"merging {self.full_name} with whith len of vrd= {len(self.vrd_0)}")
        for i, d in enumerate(self.vr_data):
            self.vr_data[i] = getattr(self, f"vrd_{i}")

        # update aliases
        self.create_alialises()
        # print(f"new length = {len(self.vr_data[0])}")

    def __read_state__(self, directory: str) -> None:
        """read virtual reservoir data from csv-file into a dataframe

        The CSV file must have the following columns

        Model Time     t
        X1
        X2

        """

        from pathlib import Path

        if self.sp.mo.register == "None":
            fn = f"{directory}/state_{self.mo.n}_vr_{self.full_name}.csv"
        elif self.sp.mo.register == "local":
            fn = f"{directory}/state_{self.full_name}.csv"
        else:
            raise ESBMTKFunctionError(
                f"Model register keyword must be 'None'/'local' not {self.sp.mo.register}"
            )

        file_path = Path(fn)

        if not file_path.exists():
            raise ESBMTKFunctionError(f"File {fn} not found")
        logging.info(f"reading state for {self.full_name} from {fn}")

        # read csv file into dataframe
        self.df: pd.DataFrame = pd.read_csv(fn)
        self.headers: list = list(self.df.columns.values)
        df = self.df
        headers = self.headers
        # print(f"reading from {fn}")
        for i, n in enumerate(headers):
            # first column is time
            if i > 0:
                # print(f"i = {i}, header = {n}, data = {df.iloc[-3:, i]}")
                self.vr_data[i - 1][:3] = df.iloc[-3:, i]

    def __sub_sample_data__(self, stride) -> None:
        """There is usually no need to keep more than a thousand data points
        so we subsample the results before saving, or processing them

        """

        # print(f"subsampling {self.fullname}")

        new: list = []
        for d in self.vr_data:
            n = d[2:-2:stride]
            new.append(n)

        self.vr_data = new
        # update aliases
        self.create_alialises()

    def __write_data__(
        self,
        prefix: str,
        start: int,
        stop: int,
        stride: int,
        append: bool,
        directory: str,
    ) -> None:
        """To be called by write_data and save_state"""

        from pathlib import Path

        p = Path(directory)
        p.mkdir(parents=True, exist_ok=True)

        mo = self.sp.mo  # model handle
        rn = self.full_name  # reservoir name
        mn = self.sp.mo.n  # model name
        mtu = f"{mo.t_unit:~P}"

        if self.sp.mo.register == "None":
            fn = f"{directory}/{prefix}{mn}_vr_{rn}.csv"  # file name
        elif self.sp.mo.register == "local":
            fn = f"{directory}/{prefix}{rn}.csv"  # file name
        else:
            raise ESBMTKFunctionError(
                f"Model register keyword must be 'None'/'local' not {self.sp.mo.register}"
            )

        df: pd.dataframe = DataFrame()

        df[f"{rn} Time [{mtu}]"] = self.mo.time[start:stop:stride]  # time

        for i, d in enumerate(self.vr_data):
            h = self.alias_list[i] if self.alias_list != "None" else f"X{i}"
            df[h] = d[start:stop:stride]

        file_path = Path(fn)
        if append and file_path.exists():
            df.to_csv(file_path, header=False, mode="a", index=False)
        else:
            df.to_csv(file_path, header=True, mode="w", index=False)
        return df


class VirtualReservoir_no_set(ExternalCode):
    """Alias to ensure backwards compatibility"""


class VirtualReservoir(Reservoir):
    """A virtual reservoir. Unlike regular reservoirs, the mass of a
    virtual reservoir depends entirely on the return value of a function.

    Example::

        VirtualReservoir(name="foo",
                    volume="10 liter",
                    concentration="1 mmol",
                    species=  ,
                    function=bar,
                    a1 to a3 =  to 3optional function arguments,
                    display_precision = number, optional, inherited from Model,
                    )

    the concentration argument will be used to initialize the reservoir and
    to determine the display units.

    The function definition follows the GenericFunction class.
    which takes a generic function and up to 6 optional
    function arguments, and will replace the mass value(s) of the
    given reservoirs with whatever the function calculates. This is
    particularly useful e.g., to calculate the pH of a given reservoir
    as function of e.g., Alkalinity and DIC.
    Parameters:
     - name = name of process,
     - act_on = name of a reservoir this process will act upon
     - function  = a function reference
     - a1 to a3 function arguments

    The function must return a list of numbers which correspond to the
    data which describe a reservoir i.e., mass, light isotope, heavy
    isotope, delta, and concentration

    In order to use this function we need first declare a function we plan to
    use with the generic function process. This function needs to follow this
    template::

        def my_func(i, a1, a2, a3) -> tuple:
            #
            # i = index of the current timestep
            # a1 to a3 =  optional function parameter. These must be present,
            # even if your function will not use it See above for details

            # calc some stuff and return it as

            return [m, l, h, d, c] # where m= mass, and l & h are the respective
                                   # isotopes. d denotes the delta value and
                                   # c the concentration
                                   # Use dummy value as necessary.

    This class provides an update method to resolve cases where e.g., two virtual
    reservoirs have a circular reference. See the documentation of update().

    """

    def __aux_inits__(self) -> None:
        """We us the regular init methods of the Reservoir Class, and extend it in this method"""

        from .processes import GenericFunction

        # if self.register != "None":
        #    self.full_name = f"{self.full_name}.{self.name}"
        name = f"{self.full_name}_generic_function".replace(".", "_")
        logging.info(f"creating {name}")

        self.gfh = GenericFunction(
            name=name,
            function=self.function,
            a1=self.a1,
            a2=self.a2,
            a3=self.a3,
            a4=self.a4,
            act_on=self,
        )

        # we only depend on the above function. so no need
        # to be in the reservoir list
        self.mo.lor.remove(self)
        # but lets keep track of  virtual reservoir in lvr.
        self.mo.lvr.append(self)
        # print(f"added {self.name} to lvr 2")

    def update(self, **kwargs) -> None:
        """This method allows to update GenericFunction parameters after the
        VirtualReservoir has been initialized. This is most useful
        when parameters have to reference other virtual reservoirs
        which do not yet exist, e.g., when two virtual reservoirs have
        a circular reference.

        Example::

            VR.update(a1=new_parameter, a2=new_parameter)

        """

        allowed_keys: list = ["a1", "a2", "a3", "a4", "a5", "a6", "volume"]
        # loop over provided kwargs
        for key, value in kwargs.items():
            if key not in allowed_keys:
                raise ValueError("you can only change a1 to a6")
            setattr(self, key, value)  # update self
            setattr(self.gfh, key, value)  # update function


class GasReservoir(ReservoirBase):
    """This object holds reservoir specific information similar to the Reservoir class

          Example::

                  Reservoir(name = "foo",     # Name of reservoir
                            species = CO2,    # Species handle
                            delta = 20,       # initial delta - optional (defaults  to 0)
                            reservoir_mass = quantity # total mass of all gases
                                             defaults to 1.833E20 mol
                            species_ppm =  number # concentration in ppm
                            plot = "yes"/"no", defaults to yes
                            plot_transform_c = a function reference, optional (see below)
                            legend_left = str, optional, useful for plot transform
                            display_precision = number, optional, inherited from Model
                            register = optional, use to register with Reservoir Group
                            isotopes = True/False otherwise use Model.m_type
                            )



    Accesing Reservoir Data:
    ~~~~~~~~~~~~~~~~~~~~~~~~

    You can access the reservoir data as:

    - Name.m # species mass
    - Name.l # mass of light isotope
    - Name.d # species delta (only avaible after M.get_delta_values()
    - Name.c # partial pressure
    - Name.v # total gas mass

    Useful methods include:

    - Name.write_data() # save data to file
    - Name.info()   # info Reservoir
    """

    def __init__(self, **kwargs) -> None:
        """Initialize a reservoir."""

        from esbmtk import Q_, Species, Model
        from typing import Callable

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "delta": [0, (int, float)],
            "reservoir_mass": ["1.833E20 mol", (str, Q_)],
            "species_ppm": ["None", (str, Q_)],
            "plot_transform_c": ["None", (str, Callable)],
            "legend_left": ["None", (str)],
            "plot": ["yes", (str)],
            "groupname": ["None", (str)],
            "function": ["None", (str, Callable)],
            "display_precision": [0.01, (int, float)],
            "register": ["None", (str, Model)],
            "full_name": ["None", (str)],
            "isotopes": [False, (bool)],
            "geometry": ["None", (str, dict)],
            "rtype": ["regular", (str)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = [
            "name",
            "species",
            "species_ppm",
            "register",
        ]

        self.__initialize_keyword_variables__(kwargs)

        self.__set_legacy_names__(kwargs)

        self.v_unit = Q_("mole").units

        # setup base data
        if isinstance(self.reservoir_mass, str):
            self.reservoir_mass = Q_(self.reservoir_mass)
        if isinstance(self.species_ppm, str):
            self.species_ppm = Q_(self.species_ppm)

        # not sure this universally true but it works for carbon
        self.species_mass = (self.reservoir_mass * self.species_ppm).to("mol")
        self.display_as = "ppm"
        self.plt_units = "ppm"

        # we use the existing approach to calculate concentration
        # which will divide species_mass/volume.
        self.volume = self.reservoir_mass
        #    Q_(self.species_mass).magnitude / self.species_ppm.to("dimensionless")
        # ).magnitude

        if self.v_unit != self.volume.units:
            raise GasReservoirError(
                f"\n\n{self.full_name} reservoir_mass units must be "
                f"in {self.v_unit} "
                f"not {self.volume.units}"
            )

        # This should probably be species specific?
        self.mu: str = "ppm"  # massunit xxxx

        # save the unit which was provided by the user for display purposes
        # left y-axis label
        self.lm: str = f"{self.species.n} [{self.mu}]"

        # initialize vectors
        self.m: NDArrayFloat = (
            np.zeros(self.species.mo.steps) + self.species_mass.magnitude
        )
        self.l: NDArrayFloat = np.zeros(self.mo.steps)
        # initialize concentration vector
        self.c: NDArrayFloat = self.m / self.volume.to(self.v_unit).magnitude
        # isotope mass
        # self.l = get_l_mass(self.m, self.delta, self.species.r)
        self.l = get_l_mass(self.c, self.delta, self.species.r)
        # delta of reservoir
        self.v: float = (
            np.zeros(self.mo.steps) + self.volume.to(self.v_unit).magnitude
        )  # mass of atmosphere

        if self.mo.number_of_solving_iterations > 0:
            self.mc = np.empty(0)
            self.cc = np.empty(0)
            self.lc = np.empty(0)
            self.vc = np.empty(0)

        self.mo.lor.append(self)  # add fthis reservoir to the model
        self.mo.lic.append(self)  # reservoir type object list
        # register instance name in global name space

        # register this group object in the global namespace
        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name_new__()

        # decide which setitem functions to use
        if self.isotopes:
            self.__set_data__ = self.__set_with_isotopes__
        else:
            self.__set_data__ = self.__set_without_isotopes__

        # any auxilliary init - normally empty, but we use it here to extend the
        # reservoir class in virtual reservoirs
        self.__aux_inits__()
        self.state = 0

    def __set_with_isotopes__(self, i: int, value: float) -> None:
        """write data by index"""

        self.m[i]: float = value[0]
        self.l[i]: float = value[1]
        # self.c[i]: float = self.m[i] / self.v[i]  # update concentration
        # self.v[i]: float = self.v[i - 1] + value[0]
        # self.c[i]: float = self.m[i] / self.v[i]  # update concentration
        # self.h[i]: float = value[2]
        # self.d[i]: float = get_delta(self.l[i], self.h[i], self.sp.r)

    def __set_without_isotopes__(self, i: int, value: float) -> None:
        """write data by index"""
        self.m[i]: float = value[0]
        # self.c[i]: float = self.m[i] / self.v[i]  # update concentration
        # self.v[i]: float = self.v[i - 1] + value[0]


class ExternalData(esbmtkBase):
    """Instances of this class hold external X/Y data which can be associated with
    a reservoir.

    Example::

           ExternalData(name       = "Name"
                        filename   = "filename",
                        legend     = "label",
                        offset     = "0 yrs",
                        reservoir  = reservoir_handle,
                        scale      = scaling factor, optional
                        display_precision = number, optional, inherited from Model
                        convert_to = optional, see below
                       )

    The data must exist as CSV file, where the first column contains
    the X-values, and the second column contains the Y-values.

    The x-values must be time and specify the time units in the header between square brackets
    They will be mapped into the model time units.

    The y-values can be any data, but the user must take care that they match the model units
    defined in the model instance. So your data file mujst look like this

    Time [years], Data [units], Data [units]
    1, 12
    2, 13

    By convention, the secon column should contaain the same type of
    data as the reservoir (i.e., a concentration), whereas the third
    column contain isotope delta values. Columns with no data should
    be left empty (and have no header!) The optional scale argument, will
    only affect the Y-col data, not the isotope data

    The column headers are only used for the time or concentration
    data conversion, and are ignored by the default plotting
    methods, but they are available as self.xh,yh

    The file must exist in the local working directory.

    the convert_to keyword can be used to force a specific conversion.
    The default is to convert into the model concentration units.

    Methods:
      - name.plot()

    Data:
      - name.x
      - name.y
      - name.df = dataframe as read from csv file

    """

    def __init__(self, **kwargs: dict[str, str]):
        from esbmtk import Q_, Model, Reservoir, DataField

        # dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "filename": ["None", (str)],
            "legend": ["None", (str)],
            "reservoir": ["None", (str, Reservoir, DataField, GasReservoir)],
            "offset": ["0 yrs", (Q_, str)],
            "display_precision": [0.01, (int, float)],
            "scale": [1, (int, float)],
            "register": ["None", (str, Model, Reservoir, DataField, GasReservoir)],
            "plot_transform_c": ["None", (str, callable)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["name", "filename", "legend", "reservoir"]

        self.__initialize_keyword_variables__(kwargs)

        # legacy names
        if self.register == "None":
            self.register = self.reservoir

        self.n: str = self.name  # string =  name of this instance
        self.fn: str = self.filename  # string = filename of data
        if isinstance(self.reservoir, Reservoir):
            self.mo: Model = self.reservoir.species.mo
        else:
            self.mo = self.reservoir.mo

        self.model = self.mo

        self.parent = self.reservoir
        self.mo.led.append(self)  # keep track of this instance

        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        if not os.path.exists(self.fn):  # check if the file is actually there
            raise ExternalDataError(f"Cannot find file {self.fn}")

        self.df: pd.DataFrame = pd.read_csv(self.fn)  # read file

        ncols = len(self.df.columns)
        if ncols < 2:  # test of we have at elast 2 columns
            raise ExternalDataError("CSV file must have at least 2 columns")
        elif ncols == 2:
            self.isotopes = False
        elif ncols == 3:
            self.isotopes = True
        elif ncols > 3:
            raise ExternalDataError("External data only supports up to 2 Y columns")
        else:
            raise ExternalDataError("ED: This should not happen")

        # print(f"Model = {self.mo.full_name}, t_unit = {self.mo.t_unit}")
        self.offset = self.ensure_q(self.offset)
        self.offset = self.offset.to(self.mo.t_unit).magnitude

        # get unit information
        self.xq = Q_(get_string_between_brackets(self.df.columns[0]))
        self.yq = Q_(get_string_between_brackets(self.df.columns[1]))
        xs = self.xq.to(self.mo.d_unit).magnitude
        ys = self.yq.to(self.register.plt_units).magnitude

        # scale input data into model  units
        self.x: NDArrayFloat = self.df.iloc[:, 0].to_numpy() * xs
        self.y: NDArrayFloat = self.df.iloc[:, 1].to_numpy() * ys

        # map into model space
        self.x = self.x - self.x[0] + self.offset

        # test for plt_transform
        if self.plot_transform_c != "None":
            if callable(self.plot_transform_c):
                self.y = self.plot_transform_c(self.y)
            else:
                raise ExternalDataError("Plot transform must be a function")

        # zh = self.df.columns[2]
        if self.isotopes:
            self.z = self.df.iloc[:, 2].to_numpy()

        # register with reservoir
        self.__register__(self.reservoir)

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name_new__()

    def __register__(self, obj):
        """Register this dataset with a flux or reservoir. This will have the
        effect that the data will be printed together with the model
        results for this reservoir

        Example::

        ExternalData.register(Reservoir)

        """
        self.obj = obj  # reser handle we associate with
        obj.led.append(self)

    def __interpolate__(self) -> None:
        """Interpolate the input data with a resolution of dt across the model
        domain The first and last data point must coincide with the
        model start and end time. In other words, this method will not
        patch data at the end points.

        This will replace the original values of name.x and name.y. However
        the original data remains accessible as name.df


        """

        xi: NDArrayFloat = self.model.time

        if (self.x[0] > xi[0]) or (self.x[-1] < xi[-1]):
            message = (
                f"\n Interpolation requires that the time domain"
                f"is equal or greater than the model domain"
                f"data t(0) = {self.x[0]}, tmax = {self.x[-1]}"
                f"model t(0) = {xi[0]}, tmax = {xi[-1]}"
            )

            raise ExternalDataError(message)
        else:
            self.y: NDArrayFloat = np.interp(xi, self.x, self.y)
            self.x = xi

    def plot(self) -> None:
        """Plot the data and save a pdf

        Example::

                ExternalData.plot()

        """

        fig, ax = plt.subplots()  #
        ax.scatter(self.x, self.y)
        ax.set_label(self.legend)
        ax.set_xlabel(self.xh)
        ax.set_ylabel(self.yh)
        plt.show()
        plt.savefig(self.n + ".pdf")



from __future__ import annotations


from numpy import array as npa
from esbmtk import carbonate_system_1_ode, carbonate_system_2_ode
from esbmtk import gas_exchange_ode, gas_exchange_ode_with_isotopes

class setup_ode():
    '''Class stub to enable state in the equation system passed to ODEINT
    '''
    

    def __init__(self, M: Model)->None:
        ''' Use this method to initialize all variables that require the state
            t-1
        '''

    def eqs(self, t, R: list, M: Model, area_table, area_dz_table, Csat_table) -> list:
        '''Auto generated esbmtk equations do not edit
        '''

 


# ---------------- write computed reservoir equations -------- #
# that do not depend on fluxes

# ---------------- write all flux equations ------------------- #
        M_C_weathering_to_sb_PO4_river__F = M.C_weathering_to_sb_PO4_river._F.rate + M.CR(t)[0]  # Signal
        M_downwelling_PO4__F = M.downwelling_PO4.scale * R[0]
        M_upwelling_PO4__F = M.upwelling_PO4.scale * R[1]
        M_C_sb_to_db_PO4_primary_production__F = M.C_sb_to_db_PO4_primary_production.scale * R[0]
        M_C_db_to_burial_PO4_burial__F = M.C_db_to_burial_PO4_burial.scale * M_C_sb_to_db_PO4_primary_production__F

# ---------------- write computed reservoir equations -------- #
# that do depend on fluxes

# ---------------- write input only reservoir equations -------- #

# ---------------- write regular reservoir equations ------------ #
        dCdt_M_sb = (
            + M_C_weathering_to_sb_PO4_river__F
            - M_downwelling_PO4__F
            + M_upwelling_PO4__F
            - M_C_sb_to_db_PO4_primary_production__F
        )/2.9999999999999996e+19

        dCdt_M_db = (
            + M_downwelling_PO4__F
            - M_upwelling_PO4__F
            + M_C_sb_to_db_PO4_primary_production__F
            - M_C_db_to_burial_PO4_burial__F
        )/9.999999999999999e+20


# ---------------- write isotope reservoir equations ------------ #

# ---------------- bits and pieces --------------------------- #
        return [
            dCdt_M_sb,  # 0
            dCdt_M_db,  # 1
        ]



"""
esbmtk: A general purpose Earth Science box model toolkit Copyright
(C), 2020 Ulrich G.  Wortmann

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import typing as tp
from esbmtk import ExternalCode
from esbmtk.experimental_reactions.bio_geochemical_reactions import (
    photosynthesis,
    OM_remineralization,
    gas_exchange_with_isotopes_2,
    gas_exchange_no_isotopes_2,
    carbonate_system_3,
)

if tp.TYPE_CHECKING:
    from esbmtk import Flux, ReservoirGroup


def init_photosynthesis(
    rg,
    productivity,
    piston_velocity,
    O2_At,
    CO2_At,
    CaCO3_reactions,
):
    """Setup photosynthesis instances"""

    M = rg.mo
    ec = ExternalCode(
        name="ps",
        species=rg.mo.Oxygen.O2,
        fname="photosynthesis",
        function=photosynthesis,
        ftype="cs2",  # cs1 is independent of fluxes, cs2 is not
        vr_datafields={
            "POM": 0.0,
            "POM_l": 0.0,
            "CO2aq": rg.swc.co2,
        },
        function_input_data=[
            O2_At,
            CO2_At,
            rg.O2,
            rg.TA,
            rg.DIC,
            rg.PO4,
            rg.SO4,
            rg.H2S,
            "Hplus",
            "CO2aq",
            productivity,
        ],
        function_params=(
            rg.volume.magnitude,
            rg.area,
            rg.sed_area,
            M.PC_ratio,
            M.NC_ratio,
            M.O2C_ratio,
            M.PUE,
            M.rain_rate,
            M.OM_frac / 1000.0 + 1.0,
            M.alpha,
            rg.swc.K1,
            rg.swc.K1K1,
            rg.swc.K1K2,
            rg.swc.KW,
            rg.swc.KB,
            rg.swc.boron,
            M.Carbon.r,
            rg.swc.p_H2O,
            piston_velocity,
            rg.swc.SA_o2,
            rg.swc.SA_co2,
            rg.swc.a_db,
            rg.swc.a_dg,
            rg.swc.a_u,
            CaCO3_reactions,
        ),
        register=rg,
        return_values=[
            {f"R_{rg.full_name}.Hplus": rg.swc.hplus},
            {f"R_{rg.full_name}.CO2aq": rg.swc.co2aq},
            {f"F_{rg.full_name}.O2": "photosynthesis"},
            {f"F_{rg.full_name}.TA": "photosynthesis"},
            {f"F_{rg.full_name}.PO4": "photosynthesis"},
            {f"F_{rg.full_name}.SO4": "photosynthesis"},
            {f"F_{rg.full_name}.H2S": "photosynthesis"},
            {f"F_{rg.full_name}.DIC": "photosynthesis"},
            {f"F_M.O2_At": "photosynthesis"},
            {f"F_M.CO2_At": "photosynthesis"},
            {f"F_{rg.full_name}.POM": "photosynthesis"},
            {f"F_{rg.full_name}.PIC": "photosynthesis"},
        ],
    )
    rg.mo.lpc_f.append(ec.fname)

    return ec


def init_OM_remineralization(
    rg: ReservoirGroup,
    pom_fluxes: list[Flux],
    pom_fluxes_l: list[Flux],
    pic_fluxes: list[Flux],
    pic_fluxes_l: list[Flux],
    pom_remin_fractions: list[float],
    pic_remin_fractions: list[float],
    CaCO3_reactions: bool,
):
    """ """

    if not isinstance(pom_remin_fractions, list):
        pom_remin_fractions = list(pom_remin_fractions)
    if not isinstance(pic_remin_fractions, list):
        pic_remin_fractions = list(pic_remin_fractions)

    M = rg.mo
    ec = ExternalCode(
        name="rm",
        species=rg.mo.Carbon.CO2,
        function=OM_remineralization,
        fname="OM_remineralization",
        ftype="cs2",  # cs1 is independent of fluxes, cs2 is not
        # hplus is not used but needed in post processing
        function_input_data=[
            pom_fluxes,
            pom_fluxes_l,
            pic_fluxes,
            pic_fluxes_l,
            pom_remin_fractions,
            pic_remin_fractions,
            rg.H2S,
            rg.SO4,
            rg.O2,
            rg.PO4,
            rg.DIC,
            rg.TA,
            "Hplus",
        ],
        function_params=(
            rg.volume.magnitude,
            rg.area,
            rg.sed_area,
            M.PC_ratio,
            M.NC_ratio,
            M.O2C_ratio,
            M.P_burial,
            M.alpha,
            rg.swc.K1,
            rg.swc.K1K1,
            rg.swc.K1K2,
            rg.swc.KW,
            rg.swc.KB,
            rg.swc.boron,
            rg.DIC.species.r,
            M.OM_d,
            CaCO3_reactions,
        ),
        register=rg,
        return_values=[
            {f"F_{rg.full_name}.DIC": "OM_remineralization"},
            {f"F_{rg.full_name}.TA": "OM_remineralization"},
            {f"F_{rg.full_name}.H2S": "OM_remineralization"},
            {f"F_{rg.full_name}.SO4": "OM_remineralization"},
            {f"F_{rg.full_name}.O2": "OM_remineralization"},
            {f"F_{rg.full_name}.PO4": "OM_remineralization"},
            {"F_M.O2_At": "OM_weathering_O2"},
            {"F_M.CO2_At": "OM_weathering_CO2"},
            {f"R_{rg.full_name}.Hplus": rg.swc.hplus},
        ],
    )
    rg.mo.lpc_f.append(ec.fname)
    return ec


def init_carbonate_system_3(
    rg: ReservoirGroup,
    pic_export_flux: Flux,
    r_sb: ReservoirGroup,
    r_ib: ReservoirGroup,
    r_db: ReservoirGroup,
    kwargs: dict,
):
    ec = ExternalCode(
        name="cs",
        species=rg.mo.Carbon.CO2,
        function=carbonate_system_3,
        fname="carbonate_system_3",
        ftype="cs2",
        r_s=r_sb,  # source (RG) of CaCO3 flux,
        r_d=r_db,  # sink (RG) of CaCO3 flux,
        function_input_data=[
            pic_export_flux,  # 1
            r_db.DIC,  # 2
            r_db.TA,  # 3
            r_sb.DIC,  # 4
            "Hplus",  # 5
            "zsnow",  # 6
            "CO3",
        ],
        function_params=(
            kwargs["Ksp0"],  # 7
            float(kwargs["kc"]),  # 8
            float(abs(kwargs["zsat0"])),  # 10
            float(kwargs["I_caco3"]),  # 11
            float(kwargs["alpha"]),  # 12
            float(abs(kwargs["zsat_min"])),  # 13
            float(abs(kwargs["zmax"])),  # 14
            float(abs(kwargs["z0"])),  # 15
            rg.swc.K2,
            rg.swc.K1K2,
            rg.swc.ca2,
            r_ib.area,  # area at top of intermediate water
            r_ib.mo.hyp.oa,  # total ocean area
            "area_table",
            "area_dz_table",
            "Csat_table",
        ),
        return_values=[
            {f"F_{rg.full_name}.DIC": "db_OM_remineralization"},
            {f"F_{rg.full_name}.TA": "db_OM_remineralization"},
            {f"R_{rg.full_name}.zsnow": float(abs(kwargs["zsnow"]))},
            {f"R_{rg.full_name}.CO3": rg.swc.co3},
        ],
        register=rg,
    )
    rg.mo.lpc_f.append(ec.fname)

    return ec


def init_gas_exchange_no_isotopes(
    r_gas,
    r_liquid,
    r_reference,
    species,
    piston_velocity,
    solubility,
):
    """Setup GasExchange instances"""

    # convert pv into model units
    pv = piston_velocity.to("meter/yr").magnitude

    ec = ExternalCode(
        name=f"{r_liquid.parent.name}_{r_liquid.name}_exchange",
        species=species,
        fname="gas_exchange_no_isotopes_2",
        function=gas_exchange_no_isotopes_2,
        ftype="cs1",  # cs1 is independent of fluxes, cs2 is not
        vr_datafields={},
        function_input_data=[
            r_gas,
            r_liquid,
        ],
        function_params=(
            r_liquid.parent.area,
            solubility,
            pv,
            r_liquid.parent.swc.p_H2O,
        ),
        register=r_gas,
        return_values=[
            {f"F_{r_gas.full_name}": "exchange"},
        ],
    )
    # print(f"return value = {ec.return_values}")
    r_liquid.mo.lpc_f.append(ec.fname)

    return ec


def init_gas_exchange_with_isotopes(
    r_gas,
    r_liquid,
    r_reference,
    species,
    piston_velocity,
    solubility,
):
    """Setup GasExchange instances"""

    # convert pv into model units
    pv = piston_velocity.to("meter/yr").magnitude

    ec = ExternalCode(
        name=f"{r_liquid.parent.name}_{r_liquid.name}_exchange",
        species=species,
        fname="gas_exchange_with_isotopes_2",
        function=gas_exchange_with_isotopes_2,
        ftype="cs1",  # cs1 is independent of fluxes, cs2 is not
        vr_datafields={},
        function_input_data=[
            r_gas,
            r_liquid,
            r_reference,
        ],
        function_params=(
            r_liquid.parent.area,
            solubility,
            pv,
            r_liquid.parent.swc.p_H2O,
            r_liquid.parent.swc.a_db,
            r_liquid.parent.swc.a_dg,
            r_liquid.parent.swc.a_u,
        ),
        register=r_gas,
        return_values=[
            {f"F_{r_gas.full_name}": "exchange"},
        ],
    )
    r_liquid.mo.lpc_f.append(ec.fname)

    return ec
